/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdlib.h>
#include <string.h>

#include <gstparseau.h>


/* elementfactory information */
static GstElementDetails gst_parseau_details = {
  ".au parser",
  "Parser/Audio",
  "Parse an .au file into raw audio",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* typefactory for 'au' */
static GstTypeFactory aufactory = {
  "audio/au",
  ".au",
  NULL,
};

/* ParseAu signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void gst_parseau_class_init(GstParseAuClass *klass);
static void gst_parseau_init(GstParseAu *parseau);

static void gst_parseau_chain(GstPad *pad,GstBuffer *buf);

static GstElementClass *parent_class = NULL;
//static guint gst_parseau_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_parseau_get_type(void) {
  static GtkType parseau_type = 0;

  if (!parseau_type) {
    static const GtkTypeInfo parseau_info = {
      "GstParseAu",
      sizeof(GstParseAu),
      sizeof(GstParseAuClass),
      (GtkClassInitFunc)gst_parseau_class_init,
      (GtkObjectInitFunc)gst_parseau_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    parseau_type = gtk_type_unique(GST_TYPE_ELEMENT,&parseau_info);
  }
  return parseau_type;
}

static void
gst_parseau_class_init(GstParseAuClass *klass) {
  GstElementClass *gstelement_class;

  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);
}

static void gst_parseau_init(GstParseAu *parseau) {
  parseau->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(parseau),parseau->sinkpad);
  gst_pad_set_chain_function(parseau->sinkpad,gst_parseau_chain);

  parseau->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(parseau),parseau->srcpad);

  parseau->offset = 0;
  parseau->size = 0;
  parseau->encoding = 0;
  parseau->frequency = 0;
  parseau->channels = 0;
}

static void gst_parseau_chain(GstPad *pad,GstBuffer *buf) {
  GstParseAu *parseau;
  gchar *data;
  glong size;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  parseau = GST_PARSEAU(pad->parent);
  g_print("gst_parseau_chain: got buffer in '%s'\n",
          gst_element_get_name(GST_ELEMENT(parseau)));
  data = (guchar *)GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  /* if we haven't seen any data yet... */
  if (parseau->size == 0) {
    GstBuffer *newbuf;
    glong *head = (glong *)data;
    /* normal format is big endian (au is a Sparc format) */
    if (GULONG_FROM_BE(*(head++)) == 0x2e736e64) {
      parseau->le = 0;
      parseau->offset = GULONG_FROM_BE(*(head++));
      parseau->size = GULONG_FROM_BE(*(head++));
      parseau->encoding = GULONG_FROM_BE(*(head++));
      parseau->frequency = GULONG_FROM_BE(*(head++));
      parseau->channels = GULONG_FROM_BE(*(head++));
      parseau->bps = ((parseau->encoding < 3) ? 1 : 2) * parseau->channels;
      g_print("offset %ld, size %ld, encoding %ld, frequency %ld, channels %ld\n",
               parseau->offset,parseau->size,parseau->encoding,
               parseau->frequency,parseau->channels);
    /* but I wouldn't be surprised by a little endian version */
    } else if (GULONG_FROM_LE(*(head++)) == 0x2e736e64) {
      parseau->le = 1;
      parseau->offset = GULONG_FROM_LE(*(head++));
      parseau->size = GULONG_FROM_LE(*(head++));
      parseau->encoding = GULONG_FROM_LE(*(head++));
      parseau->frequency = GULONG_FROM_LE(*(head++));
      parseau->channels = GULONG_FROM_LE(*(head++));
      parseau->bps = ((parseau->encoding < 3) ? 1 : 2) * parseau->channels;
      g_print("offset %ld, size %ld, encoding %ld, frequency %ld, channels %ld\n",
               parseau->offset,parseau->size,parseau->encoding,
               parseau->frequency,parseau->channels);
    } else {
      g_print("help, dunno what I'm looking at!\n");
      gst_buffer_unref(buf);
      return;
    }

    switch (parseau->encoding) {
      case 1:
        parseau->meta.format = AFMT_MU_LAW;break;
      case 2:
        parseau->meta.format = AFMT_S8;break;
      case 3:
        parseau->meta.format = AFMT_S16_LE;break;
      default:
        g_print("help!, don't know how to deal with this format yet\n");
        return;
    }
    parseau->meta.frequency = parseau->frequency;
    parseau->meta.channels = parseau->channels;

    newbuf = GST_BUFFER(gst_buffer_new());
    GST_BUFFER_DATA(newbuf) = (gpointer)malloc(size-(parseau->offset));
    memcpy(GST_BUFFER_DATA(newbuf),data+24,size-(parseau->offset));
    gst_buffer_unref(buf);
    GST_BUFFER_SIZE(newbuf) = size-(parseau->offset);
//    FIXME
//    newbuf->meta = (gpointer)malloc(sizeof(MetaAudioRaw));
//    memcpy(newbuf->meta,&parseau->meta,sizeof(parseau->meta));
//    ((MetaAudioRaw *)newbuf->meta)->samples = (size-(parseau->offset)) /
//parseau->bps;
    gst_pad_push(parseau->srcpad,newbuf);
    return;
  }


//  FIXME
//  buf->meta = (gpointer)malloc(sizeof(MetaAudioRaw));
//  memcpy(buf->meta,&parseau->meta,sizeof(parseau->meta));
//  ((MetaAudioRaw *)buf->meta)->samples = size / parseau->bps;

  gst_pad_push(parseau->srcpad,buf);
}


GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* create the plugin structure */
  plugin = gst_plugin_new ("parseau");
  g_return_val_if_fail (plugin != NULL, NULL);

  /* create an elementfactory for the parseau element and list it */
  factory = gst_elementfactory_new ("parseau",GST_TYPE_PARSEAU,
                                    &gst_parseau_details);
  g_return_val_if_fail (factory != NULL, NULL);
  gst_plugin_add_factory (plugin,factory);

  return plugin;
}
