/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */



#include <stdlib.h>
#include <string.h>

//#define DEBUG_ENABLED
#include <gstaviencoder.h>



/* elementfactory information */
GstElementDetails gst_aviencoder_details = {
  ".avi encoder",
  "Encoder/Video",
  "Encodes audio and video into an avi stream",
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

static GstCaps*  avi_typefind(GstBuffer *buf,gpointer private);

/* typefactory for 'avi' */
GstTypeFactory avifactory = {
  "video/avi",
  ".avi",
  avi_typefind,
};


/* AviEncoder signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_VIDEO,
  ARG_AUDIO,
  /* FILL ME */
};

static void gst_aviencoder_class_init(GstAviEncoderClass *klass);
static void gst_aviencoder_init(GstAviEncoder *aviencoder);

static void gst_aviencoder_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_aviencoder_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstElementClass *parent_class = NULL;
//static guint gst_aviencoder_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_aviencoder_get_type(void) {
  static GtkType aviencoder_type = 0;

  if (!aviencoder_type) {
    static const GtkTypeInfo aviencoder_info = {
      "GstAviEncoder",
      sizeof(GstAviEncoder),
      sizeof(GstAviEncoderClass),
      (GtkClassInitFunc)gst_aviencoder_class_init,
      (GtkObjectInitFunc)gst_aviencoder_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    aviencoder_type = gtk_type_unique(GST_TYPE_ELEMENT,&aviencoder_info);
  }
  return aviencoder_type;
}

static void
gst_aviencoder_class_init(GstAviEncoderClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  gtk_object_add_arg_type("GstAviEncoder::video", GTK_TYPE_STRING,
                           GTK_ARG_READWRITE, ARG_VIDEO);
  gtk_object_add_arg_type("GstAviEncoder::audio", GTK_TYPE_STRING,
                           GTK_ARG_READWRITE, ARG_AUDIO);

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_aviencoder_set_arg;
  gtkobject_class->get_arg = gst_aviencoder_get_arg;

}

static void gst_aviencoder_init(GstAviEncoder *aviencoder) {
  aviencoder->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(aviencoder),aviencoder->srcpad);

  aviencoder->state = GST_AVIENCODER_INITIAL;
  aviencoder->riff = NULL;
  aviencoder->num_audio_pads = 0;
  aviencoder->num_video_pads = 0;
  aviencoder->next_time = 0;

  aviencoder->riff = gst_riff_encoder_new(GST_RIFF_RIFF_AVI);
  aviencoder->aviheader = g_malloc0(sizeof(gst_riff_avih));
}

GstElement *gst_aviencoder_new(gchar *name) {
  GstElement *aviencoder = GST_ELEMENT(gtk_type_new(GST_TYPE_AVIENCODER));
  gst_element_set_name(GST_ELEMENT(aviencoder),name);
  return aviencoder;
}

static GstCaps* 
avi_typefind (GstBuffer *buf, gpointer private) 
{
  gchar *data = GST_BUFFER_DATA(buf);
  GstCaps *new;

  GST_DEBUG (0,"parseavi: typefind\n");
  if (strncmp(&data[0], "RIFF", 4)) return NULL;
    if (strncmp(&data[8], "AVI ", 4)) return NULL;

  new = gst_caps_new ("avi_typefind", "video/avi");

  return new;
}

static void gst_aviencoder_make_header(GstAviEncoder *aviencoder) 
{
  int i;

  gst_riff_strh strh;

  aviencoder->aviheader->us_frame = 40000;
  aviencoder->aviheader->streams  = aviencoder->num_video_pads + aviencoder->num_audio_pads;
  aviencoder->aviheader->width    = -1;
  aviencoder->aviheader->height   = -1;
  gst_riff_encoder_avih(aviencoder->riff, aviencoder->aviheader, sizeof(gst_riff_avih)); 

  memset(&strh, 0, sizeof(gst_riff_strh));
  strh.scale = 40000;
  
  gst_riff_encoder_strh(aviencoder->riff, GST_RIFF_FCC_vids, &strh, sizeof(gst_riff_strh)); 

  for (i=0; i<aviencoder->num_video_pads; i++) {
    gst_riff_encoder_strf(aviencoder->riff, aviencoder->video_header[i], sizeof(gst_riff_strf_vids)); 
  }

  /*
  gst_riff_encoder_strh(aviencoder->riff, GST_RIFF_FCC_auds, &strh, sizeof(gst_riff_strh)); 
  gst_riff_encoder_strf(aviencoder->riff, &strf_auds, sizeof(gst_riff_strf_auds)); 
  */

}

static void gst_aviencoder_chain(GstPad *pad,GstBuffer *buf) {
  GstAviEncoder *aviencoder;
  guchar *data;
  gulong size;
  const gchar *padname;
  gint channel;
  GstBuffer *newbuf;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  g_return_if_fail(GST_BUFFER_DATA(buf) != NULL);

  aviencoder = GST_AVIENCODER(pad->parent);
  data = (guchar *)GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  switch(aviencoder->state) {
    case GST_AVIENCODER_INITIAL:
      GST_DEBUG (0,"gst_aviencoder_chain: writing header\n");
      gst_aviencoder_make_header(aviencoder);
      newbuf = gst_riff_encoder_get_and_reset_buffer(aviencoder->riff);
      gst_pad_push(aviencoder->srcpad, newbuf);
      aviencoder->state = GST_AVIENCODER_MOVI;
    case GST_AVIENCODER_MOVI:
      padname = gst_pad_get_name(pad);

      if (strncmp(padname, "audio_", 6) == 0) {
        channel = atoi(&padname[6]);
        GST_DEBUG (0,"gst_aviencoder_chain: got audio buffer in from channel %02d %lu\n", channel, size);
        gst_riff_encoder_chunk(aviencoder->riff, GST_RIFF_01wb, NULL, size); 
        newbuf = gst_riff_encoder_get_and_reset_buffer(aviencoder->riff);
        gst_pad_push(aviencoder->srcpad, newbuf);
      }
      else if (strncmp(padname, "video_", 6) == 0) {
        channel = atoi(&padname[6]);
        GST_DEBUG (0,"gst_aviencoder_chain: got video buffer in from channel %02d %lu\n", channel, size);
        gst_riff_encoder_chunk(aviencoder->riff, GST_RIFF_00db, NULL, size); 
        newbuf = gst_riff_encoder_get_and_reset_buffer(aviencoder->riff);
        GST_DEBUG (0,"gst_aviencoder_chain: encoded %u\n", GST_BUFFER_SIZE(newbuf));
        gst_pad_push(aviencoder->srcpad, newbuf);
      }
      GST_BUFFER_SIZE(buf) = (GST_BUFFER_SIZE(buf)+1)&~1;
      gst_pad_push(aviencoder->srcpad, buf);
      break;
    default:
      break;
  }

  //gst_buffer_unref(buf);
}

static void gst_aviencoder_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstAviEncoder *src;
  guchar *padname;
  guchar name[20];
  guchar channel[3];
  guint channelnr;
  guint i;
  GstPad *newpad;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_AVIENCODER(object));
  src = GST_AVIENCODER(object);

  switch(id) {
    case ARG_VIDEO:
      padname = GTK_VALUE_STRING(*arg);
      channel[0] = padname[0];
      channel[1] = padname[1];
      channel[2] = 0;
      channelnr = atoi(channel);
      GST_DEBUG (0,"gst_aviencoder_set_arg: making video_%02d pad %s\n", channelnr, &padname[3]);
      i = src->num_video_pads;
      g_snprintf(name, 20, "video_%02d", i);
      newpad = gst_pad_new(name,GST_PAD_SINK);
      src->video_pad[i] = newpad;
      gst_element_add_pad(GST_ELEMENT(src),newpad);
      gst_pad_set_chain_function(newpad,gst_aviencoder_chain);
      src->video_header[i] = gst_aviencoder_make_video_header(gst_riff_fourcc_to_id(&padname[3]), 640, 480, 1, 24);
      src->num_video_pads++;
      break;
    case ARG_AUDIO:
      padname = GTK_VALUE_STRING(*arg);
      channel[0] = padname[0];
      channel[1] = padname[1];
      channel[2] = 0;
      channelnr = atoi(channel);
      GST_DEBUG (0,"gst_aviencoder_set_arg: making audio_%02d pad %s\n", channelnr, &padname[3]);
      i = src->num_audio_pads;
      g_snprintf(name, 20, "audio_%02d", i);
      newpad = gst_pad_new(name,GST_PAD_SINK);
      src->audio_pad[i] = newpad;
      gst_element_add_pad(GST_ELEMENT(src),newpad);
      gst_pad_set_chain_function(newpad,gst_aviencoder_chain);
      src->num_audio_pads++;
      break;
      break;
    default:
      break;
  }
}

static void gst_aviencoder_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstAviEncoder *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_AVIENCODER(object));
  src = GST_AVIENCODER(object);

  switch(id) {
    case ARG_VIDEO:
    break;
  default:
    arg->type = GTK_TYPE_INVALID;
    break;
  }
}

GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* this filter needs the riff parser */
  if (!gst_library_load("gstriff")) {
    gst_info("aviencoder: could not load support library: 'gstriff'\n");
    return NULL;
  }

  /* create the plugin structure */
  plugin = gst_plugin_new("aviencoder");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the aviencoder element */
  factory = gst_elementfactory_new("aviencoder",GST_TYPE_AVIENCODER,
                                   &gst_aviencoder_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
