/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_AVIENCODER_H__
#define __GST_AVIENCODER_H__


#include <config.h>
#include <gst/gst.h>
#include <gst/meta/audioraw.h>
#include <gst/meta/videoraw.h>
#include <gstriff.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GST_TYPE_AVIENCODER \
  (gst_aviencoder_get_type())
#define GST_AVIENCODER(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_AVIENCODER,GstAviEncoder))
#define GST_AVIENCODER_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_AVIENCODER,GstAviEncoder))
#define GST_IS_AVIENCODER(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_AVIENCODER))
#define GST_IS_AVIENCODER_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_AVIENCODER))


#define GST_AVIENCODER_INITIAL	  0	/* initialized state */
#define GST_AVIENCODER_MOVI	  1	/* encoding movi */

#define GST_AVIENCODER_MAX_AUDIO_PADS	8	
#define GST_AVIENCODER_MAX_VIDEO_PADS	8	

typedef struct _GstAviEncoder GstAviEncoder;
typedef struct _GstAviEncoderClass GstAviEncoderClass;

struct _GstAviEncoder {
  GstElement element;

  /* pads */
  GstPad *srcpad;

  /* AVI encoding state */
  gint state;

  /* RIFF encoding state */
  GstRiff *riff;

  guint64 next_time;
  guint64 time_interval;

  gst_riff_avih  *aviheader;    /* the avi header */
  guint num_audio_pads;
  guint num_video_pads;

  GstPad             *audio_pad[GST_AVIENCODER_MAX_AUDIO_PADS];
  MetaAudioRaw       *audio_meta[GST_AVIENCODER_MAX_AUDIO_PADS];
  gst_riff_strf_auds *audio_header[GST_AVIENCODER_MAX_AUDIO_PADS];
  
  GstPad             *video_pad[GST_AVIENCODER_MAX_VIDEO_PADS];
  MetaVideoRaw       *video_meta[GST_AVIENCODER_MAX_VIDEO_PADS];
  gst_riff_strf_vids *video_header[GST_AVIENCODER_MAX_VIDEO_PADS];
};

struct _GstAviEncoderClass {
  GstElementClass parent_class;
};

GtkType gst_aviencoder_get_type(void);
GstElement *gst_aviencoder_new(gchar *name);

gst_riff_strf_vids *gst_aviencoder_make_video_header(guint32 codec, guint width, guint height, guint planes, guint bit_cnt);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_AVIENCODER_H__ */
