/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>

//#define DEBUG_ENABLED
#define WIN32

#include "config.h"
#ifdef HAVE_LIBMMX
#include "mmx.h"
#endif

#include "gstwindec.h"
#include <libs/riff/gstriff.h>

#include <wine/driver.h>
#include <wine/vfw.h>

/* elementfactory information */
GstElementDetails gst_windec_details = {
  "Windows codec decoder",
  "Filter/Decoder/Image",
  NULL,
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

extern GstPadTemplate *src_temp;

/* Winec signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_AVI_HEADER,
  ARG_BITDEPTH,
  /* FILL ME */
};

static void gst_windec_class_init(GstWinDec *klass);
static void gst_windec_init(GstWinDec *windec);

static void gst_windec_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_windec_get_arg(GtkObject *object,GtkArg *arg,guint id);

typedef struct _GstBITMAPINFO GstBITMAPINFO;

struct _GstBITMAPINFO {
  BITMAPINFOHEADER bh;
  guint16 red_mask;
  guint16 green_mask;
  guint16 blue_mask;
};

typedef struct _GstWinLoaderData GstWinLoaderData;

struct _GstWinLoaderData {
  COMPVARS m_cv;
  GstBITMAPINFO m_obh;
  GstBITMAPINFO m_bh;
  gint bpp;
  guchar uncompr[1280000];
};


static GstElementClass *parent_class = NULL;
//static guint gst_windec_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_windec_get_type(void) {
  static GtkType windec_type = 0;

  if (!windec_type) {
    static const GtkTypeInfo windec_info = {
      "GstWinDec",
      sizeof(GstWinDec),
      sizeof(GstWinDec),
      (GtkClassInitFunc)gst_windec_class_init,
      (GtkObjectInitFunc)gst_windec_init,
      (GtkArgSetFunc)gst_windec_set_arg,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    windec_type = gtk_type_unique(GST_TYPE_ELEMENT,&windec_info);
  }
  return windec_type;
}

static void
gst_windec_class_init(GstWinDec *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstWinDec::avi_header", GTK_TYPE_POINTER,
                               GTK_ARG_WRITABLE, ARG_AVI_HEADER);
  gtk_object_add_arg_type("GstWinDec::bitdepth", GTK_TYPE_INT,
                               GTK_ARG_READWRITE, ARG_BITDEPTH);
      
  gtkobject_class->set_arg = gst_windec_set_arg;
  gtkobject_class->get_arg = gst_windec_get_arg;
}

static void gst_windec_init(GstWinDec *windec) {
  GST_DEBUG (0,"gst_windec_init: initializing\n");
  /* create the sink and src pads */
  windec->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(windec),windec->sinkpad);
  gst_pad_set_chain_function(windec->sinkpad,gst_windec_chain);
  windec->srcpad = gst_pad_new_from_template(src_temp, "src");
  gst_element_add_pad(GST_ELEMENT(windec),windec->srcpad);

  /* initialize the windec decoder state */
  windec->next_time = 0;

  // reset the initial video state
  windec->format = -1;
  windec->width = -1;
  windec->height = -1;

  windec->inited = FALSE;
  windec->prev = NULL;
  windec->bitdepth = 16;

  windec->extra_data = g_malloc0(sizeof(GstWinLoaderData));
}

GstElement *gst_windec_new(gchar *name) {
  GstElement *windec = GST_ELEMENT(gtk_type_new(GST_TYPE_WINDEC));
  gst_element_set_name(GST_ELEMENT(windec),name);
  return windec;
}

static gboolean gst_windec_init_decoder(GstWinDec *windec) 
{
  HRESULT h;
  GstBITMAPINFO temp;
  GstWinLoaderData *data;

  data = (GstWinLoaderData *)windec->extra_data;

  data->m_obh.bh.biSize=sizeof(data->m_obh);
  data->m_cv.cbSize=sizeof(data->m_cv);
  gst_info("gstwindec: trying to open library %p\n", data);
  data->m_cv.hic=ICOpen(
           GST_RIFF_FCC_vidc,
           data->m_bh.bh.biCompression,
           ICMODE_DECOMPRESS);
  if(data->m_cv.hic==0)
  {
     printf("gstwindec:: ICOpen failed\n");
     return FALSE;
  }
  data->m_cv.fccType=GST_RIFF_FCC_vidc;
  data->m_cv.fccHandler=data->m_bh.bh.biCompression,

  h=ICDecompressGetFormat(data->m_cv.hic, &data->m_bh, &data->m_obh);
  if(h!=0)
  {
    printf("gstwindec:: Can't handle this format\n");
    ICClose(data->m_cv.hic);
    return FALSE;
  }
  printf("gstwindec:: output format %08lx (%4.4s) %ld\n", data->m_obh.bh.biCompression, (gchar *)&data->m_obh.bh.biCompression, data->m_obh.bh.biHeight);

  memcpy(&temp, &data->m_obh, sizeof(data->m_obh));
  temp.bh.biBitCount=windec->bitdepth;
  temp.bh.biCompression=GST_RIFF_rgb;
  temp.red_mask = 0xF800;
  temp.green_mask = 0x07E0;
  temp.blue_mask = 0x001F;

  data->bpp = (temp.bh.biBitCount+4)/8;
  temp.bh.biSizeImage=data->m_obh.bh.biWidth*data->m_obh.bh.biHeight*data->bpp;
  h=ICDecompressQuery(data->m_cv.hic, &data->m_bh, &temp);
  if(h==0) {
    memcpy(&data->m_obh, &temp, sizeof(data->m_obh));
    printf("gstwindec:: Set Bit Depth to %d\n", temp.bh.biBitCount);
  }
  else {
    printf("gstwindec:: Set Bit Depth failed\n");
  }

  h=ICDecompressBegin(data->m_cv.hic, &data->m_bh.bh, &data->m_obh.bh);
  if(h!=0)
  {
    printf("gstwindec:: ICDecompressBegin failed\n");
    return FALSE;
  }

  windec->meta = gst_meta_new_size(sizeof(MetaVideoRaw));
  switch(data->bpp) {
    case 2:
      windec->format = ((MetaVideoRaw *)windec->meta)->format = GST_COLORSPACE_BGR565;
      break;
    case 3:
      windec->format = ((MetaVideoRaw *)windec->meta)->format = GST_COLORSPACE_BGR24;
      break;
    case 4:
      windec->format = ((MetaVideoRaw *)windec->meta)->format = GST_COLORSPACE_BGR32;
      break;
    default:
      break;
  }
  ((MetaVideoRaw *)windec->meta)->width = data->m_bh.bh.biWidth;
  ((MetaVideoRaw *)windec->meta)->height = data->m_bh.bh.biHeight;

  return TRUE;
}

#ifdef HAVE_LIBMMX
static mmx_t MMX_FFE0           = (mmx_t)(long long)0xFFE0FFE0FFE0FFE0;                
#endif

void gst_windec_chain(GstPad *pad, GstBuffer *buf) 
{
  GstWinDec *windec;
  GstWinLoaderData *data;
  GstBuffer *outbuf;
  guint32 *outdata;
  guint32 *ptr, *outptr;
  gint i, j;
  gint w;

  g_return_if_fail(buf != NULL);
  if (GST_BUFFER_DATA(buf) == NULL) return;

  windec = GST_WINDEC(pad->parent);

  if (!GST_PAD_CONNECTED(windec->srcpad)) {
    gst_buffer_unref(buf);
    return;
  }

  if (!windec->inited) {
    if (!gst_windec_init_decoder(windec)) {
      g_print("gst_windec: could not init decoder\n");
      return;
    }
    windec->inited = TRUE;
  }

  data = (GstWinLoaderData *)windec->extra_data;

  outbuf = gst_buffer_new();
  outdata = (guint32 *)GST_BUFFER_DATA(outbuf) = g_malloc(data->m_obh.bh.biSizeImage);
  GST_BUFFER_SIZE(outbuf) = data->m_obh.bh.biSizeImage;

  GST_DEBUG (0,"gst_windec: got buffer %08lx %p %u\n", *(gulong *)GST_BUFFER_DATA(buf), data, GST_BUFFER_SIZE(buf));
  
  ICDecompress(data->m_cv.hic, 0&ICDECOMPRESS_NOTKEYFRAME, &data->m_bh.bh, GST_BUFFER_DATA(buf), &data->m_obh.bh, &data->uncompr[0]);

  GST_DEBUG (0, "gst_windec: decompressed buffer %d %ld %ld\n",
             data->m_obh.bh.biBitCount, data->m_obh.bh.biWidth,
             data->m_obh.bh.biHeight);

  GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf);

  if (windec->meta) {
    gst_buffer_add_meta(outbuf, (GstMeta *)windec->meta);
    gst_meta_ref((GstMeta *)windec->meta);
  }
  
  ptr = (guint32 *) &data->uncompr[0];
  outptr = (guint32 *) (outdata + (data->m_obh.bh.biWidth * (data->m_obh.bh.biHeight-1) * data->bpp)/4);
  w = (data->m_obh.bh.biWidth * data->bpp)/2;

  if (windec->format == GST_COLORSPACE_BGR565) {
#ifdef HAVE_LIBMMX
    movq_m2r(MMX_FFE0, mm3);
#endif
    for (i=data->m_obh.bh.biHeight; i ; i--) {
      for (j=(w)/4; j ; j--) {
#ifdef HAVE_LIBMMX
        movq_m2r(*ptr, mm0);
        movq_r2r(mm0, mm2);
        pand_r2r(mm3, mm2);
        paddw_r2r(mm2, mm0);
        movq_r2m(mm0, *outptr);
        ptr+=2;
        outptr+=2;
#else
        *outptr++ = *ptr + (*ptr++&0xFFE0FFE0);
        *outptr++ = *ptr + (*ptr++&0xFFE0FFE0);
#endif
      }
      outptr -= w;
    }
  }
  else {
    for (i=data->m_obh.bh.biHeight; i ; i--) {
      for (j=(w)/4; j ; j--) {
#ifdef HAVE_LIBMMX
        movq_m2r(*ptr, mm0);
        movq_r2m(mm0, *outptr);
        ptr+=2;
        outptr+=2;
#else
        *outptr++ = ptr++;
        *outptr++ = ptr++;
#endif
      }
      outptr -= w;
    }
  }
#ifdef HAVE_LIBMMX
  emms();
#endif

  //if (windec->prev) gst_buffer_unref(windec->prev);
  //windec->prev = outbuf;
  //GST_BUFFER_FLAG_SET(outbuf, GST_BUFFER_DONTFREE);

  GST_DEBUG (0,"gst_windec: pushing buffer %llu\n", GST_BUFFER_TIMESTAMP(outbuf));
  gst_pad_push(windec->srcpad, outbuf);
  GST_DEBUG (0,"gst_windec: pushed buffer\n");
  gst_buffer_unref(buf);
}

static void gst_windec_set_arg(GtkObject *object,GtkArg *arg,guint id)
{
  GstWinDec *windec;
  GstWinLoaderData *data;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_WINDEC(object));
  windec = GST_WINDEC(object);

  data = (GstWinLoaderData *)windec->extra_data;

  switch(id) {
    case ARG_AVI_HEADER:
      memcpy(&data->m_bh, GTK_VALUE_POINTER(*arg), sizeof(BITMAPINFOHEADER));
      break;
    case ARG_BITDEPTH:
      windec->bitdepth = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void gst_windec_get_arg(GtkObject *object,GtkArg *arg,guint id)
{
  GstWinDec *windec;
  GstWinLoaderData *data;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_WINDEC(object));
  windec = GST_WINDEC(object);

  data = (GstWinLoaderData *)windec->extra_data;

  switch(id) {
    case ARG_BITDEPTH:
      GTK_VALUE_INT(*arg) = data->m_obh.bh.biBitCount;
      break;
    default:
      break;
  }
}

