/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>

//#define DEBUG_ENABLED
#include <gstv4lsrc.h>

#include <linux/videodev.h>

GstElementDetails gst_v4lsrc_details = {
  "Video (v4l) Source",
  "Source/Video",
  "Read from a Video for Linux capture device",
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

/* V4lSrc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  ARG_FORMAT,
  ARG_TUNE,
  ARG_TUNED,
  ARG_INPUT,
  ARG_NORM,
  ARG_VOLUME,
  ARG_MUTE,
  ARG_AUDIO_MODE,
  ARG_COLOR,
  ARG_BRIGHT,
  ARG_HUE,
  ARG_CONTRAST,
};


static void gst_v4lsrc_class_init(GstV4lSrcClass *klass);
static void gst_v4lsrc_init(GstV4lSrc *v4lsrc);
static void gst_v4lsrc_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_v4lsrc_get_arg(GtkObject *object,GtkArg *arg,guint id);
static GstElementStateReturn gst_v4lsrc_change_state(GstElement *element);
static void gst_v4lsrc_close_v4l(GstV4lSrc *src);
static gboolean gst_v4lsrc_open_v4l(GstV4lSrc *src);
void gst_v4lsrc_sync_parms(GstV4lSrc *v4lsrc);
void gst_v4lsrc_sync_overlay(GtkWidget *src, MetaOverlay *raw, GstV4lSrc *v4lsrc);


static GstElementClass *parent_class = NULL;
////static guint gst_v4lsrc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_v4lsrc_get_type(void) {
  static GtkType v4lsrc_type = 0;

  if (!v4lsrc_type) {
    static const GtkTypeInfo v4lsrc_info = {
      "GstV4lSrc",
      sizeof(GstV4lSrc),
      sizeof(GstV4lSrcClass),
      (GtkClassInitFunc)gst_v4lsrc_class_init,
      (GtkObjectInitFunc)gst_v4lsrc_init,
      (GtkArgSetFunc)gst_v4lsrc_set_arg,
      (GtkArgGetFunc)gst_v4lsrc_get_arg,
      (GtkClassInitFunc)NULL,
    };
    v4lsrc_type = gtk_type_unique(GST_TYPE_ELEMENT,&v4lsrc_info);
  }
  return v4lsrc_type;
}

static void
gst_v4lsrc_class_init(GstV4lSrcClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstV4lSrc::width", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_WIDTH);
  gtk_object_add_arg_type("GstV4lSrc::height", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_HEIGHT);
  gtk_object_add_arg_type("GstV4lSrc::format", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_FORMAT);

  gtk_object_add_arg_type("GstV4lSrc::tune", GTK_TYPE_ULONG,
                          GTK_ARG_WRITABLE, ARG_TUNE);
  gtk_object_add_arg_type("GstV4lSrc::tuned", GTK_TYPE_BOOL,
                          GTK_ARG_READABLE, ARG_TUNED);
  gtk_object_add_arg_type("GstV4lSrc::input", GTK_TYPE_INT,
                          GTK_ARG_WRITABLE, ARG_INPUT);
  gtk_object_add_arg_type("GstV4lSrc::norm", GTK_TYPE_INT,
                          GTK_ARG_WRITABLE, ARG_NORM);
  gtk_object_add_arg_type("GstV4lSrc::volume", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_VOLUME);
  gtk_object_add_arg_type("GstV4lSrc::mute", GTK_TYPE_BOOL,
                          GTK_ARG_READWRITE, ARG_MUTE);
  gtk_object_add_arg_type("GstV4lSrc::mode", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_AUDIO_MODE);
  gtk_object_add_arg_type("GstV4lSrc::color", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_COLOR);
  gtk_object_add_arg_type("GstV4lSrc::bright", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_BRIGHT);
  gtk_object_add_arg_type("GstV4lSrc::hue", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_HUE);
  gtk_object_add_arg_type("GstV4lSrc::contrast", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_CONTRAST);

  gtkobject_class->set_arg = gst_v4lsrc_set_arg;
  gtkobject_class->get_arg = gst_v4lsrc_get_arg;


  //gstsrc_class->push = gst_v4lsrc_push;
}

static void gst_v4lsrc_init(GstV4lSrc *v4lsrc) {
  v4lsrc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(v4lsrc),v4lsrc->srcpad);

  /* if the destination cannot say what it wants, we give this */
  v4lsrc->width = 100;
  v4lsrc->height = 100;
  v4lsrc->format = GST_COLORSPACE_RGB565;
  v4lsrc->buffer_size = v4lsrc->width * v4lsrc->height * 3;
  v4lsrc->new_meta = TRUE;
  // make a grbber
  v4lsrc->grabber = grab_init();
  v4lsrc->overlay_ok = FALSE;
  v4lsrc->overlay_disable = FALSE;
}

GstElement *gst_v4lsrc_new(gchar *name) {
  GstElement *v4lsrc = GST_ELEMENT(gtk_type_new(GST_TYPE_V4LSRC));
  gst_element_set_name(GST_ELEMENT(v4lsrc),name);
  return v4lsrc;
}


void gst_v4lsrc_push(GstElement *src) {
  GstV4lSrc *v4lsrc;
  GstBuffer *buf = NULL;
  GstMeta *meta;
  MetaVideoRaw *meta_pull;
  guint8 *grab_buf;
  gboolean validbuffer_received = FALSE;

  g_return_if_fail(src != NULL);
  g_return_if_fail(GST_IS_V4LSRC(src));
  v4lsrc = GST_V4LSRC(src);

  if (v4lsrc->in_overlay) {
    gtk_main();
    return;
  }

  // find out what the destination wants
  // and try to pull a buffer from the destination
  if (GST_PAD_CAN_PULL(v4lsrc->srcpad) && (buf = gst_pad_pull(v4lsrc->srcpad)) != NULL) {

    GST_DEBUG (0,"v4lsrc: received buffer %p\n", GST_BUFFER_DATA(buf));
    /* assume this is a valid buffer */
    validbuffer_received = TRUE;
    /* see what type it is */
    v4lsrc->meta = gst_buffer_get_first_meta(buf);
    meta_pull = (MetaVideoRaw *) gst_buffer_get_first_meta(buf);
    if (meta_pull) {
      // is is something new
      if (meta_pull->overlay_info && meta_pull->dga_info && !v4lsrc->overlay_disable) {

       	GST_DEBUG (0,"v4lsrc: dest can do overlay %p\n", v4lsrc);
        if (!v4lsrc->overlay_ok) {
          if (v4lsrc->grabber->grab_setupfb(v4lsrc->grabber, 
    	  	          	meta_pull->dga_info->swidth,
				meta_pull->dga_info->sheight,
				VIDEO_RGB16_LE,
				meta_pull->dga_info->base, 
				meta_pull->dga_info->bytes_per_line) != -1) {
	    gtk_signal_connect(GTK_OBJECT(meta_pull->overlay_info->overlay_element),"overlay-change-notify",
			     GTK_SIGNAL_FUNC(gst_v4lsrc_sync_overlay),v4lsrc);

	    v4lsrc->overlay_disable = FALSE;
	    v4lsrc->overlay_ok = TRUE;
	    v4lsrc->in_overlay = TRUE;
	    meta_pull->overlay_info->did_overlay = TRUE;
	    GST_DEBUG (0,"v4lsrc: pushing overlay notification\n");
	    gst_pad_push(v4lsrc->srcpad,buf);
	    return;
	  }
	  else {
	    v4lsrc->overlay_disable = TRUE;
	    v4lsrc->overlay_ok = TRUE;
	    v4lsrc->in_overlay = FALSE;
	    meta_pull->overlay_info->did_overlay = FALSE;
	  }
	}
	else {
	  g_assert_not_reached();
	}
      }
      if ((meta_pull->format != v4lsrc->format) ||
          (meta_pull->width != v4lsrc->width) ||
	  (meta_pull->height != v4lsrc->height)) {

        GST_DEBUG (0,"v4lsrc: received different meta %d %d\n", meta_pull->width, meta_pull->height);

        if (meta_pull->width == -1) {
	  validbuffer_received = FALSE;
	}
	else v4lsrc->width = meta_pull->width;

        if (meta_pull->height == -1){
	  validbuffer_received = FALSE;
	}
	else v4lsrc->height = meta_pull->height;


	if (validbuffer_received || meta_pull->format != v4lsrc->format) {
          GST_DEBUG (0,"v4lsrc: valid buffer\n");
          v4lsrc->format = meta_pull->format;
	  // set ourselves to the new values
	  gst_v4lsrc_sync_parms(v4lsrc);
	}
	v4lsrc->new_meta = TRUE;

	/*
        meta_pull->format = v4lsrc->format;
        meta_pull->width = v4lsrc->width;
        meta_pull->height = v4lsrc->height;
	*/
      }
      if (meta_pull->overlay_info) meta_pull->overlay_info->did_overlay = FALSE;
    }
  }

  if (!validbuffer_received) {
    buf = gst_buffer_new();
    GST_BUFFER_DATA(buf) = g_malloc(v4lsrc->buffer_size);
    GST_BUFFER_SIZE(buf) = v4lsrc->buffer_size;
    v4lsrc->new_meta = TRUE;
    GST_DEBUG (0,"v4lsrc: making new buffer %p\n", GST_BUFFER_DATA(buf));
  }
		
  if (v4lsrc->new_meta) {
    GST_DEBUG (0,"v4lsrc: making new meta\n");
    meta = gst_meta_new_size(sizeof(MetaVideoRaw));

    ((MetaVideoRaw *)meta)->format = v4lsrc->format;
    ((MetaVideoRaw *)meta)->width = v4lsrc->width;
    ((MetaVideoRaw *)meta)->height = v4lsrc->height;

    gst_buffer_add_meta(buf, meta);
    v4lsrc->new_meta = FALSE;
  }

  // request a buffer from the grabber
  grab_buf = v4lsrc->grabber->grab_capture(v4lsrc->grabber, 0);
  //meta_pull->overlay_info->did_overlay = FALSE;

  g_assert(buf != NULL);

  if (grab_buf) {
    GST_DEBUG (0,"v4lsrc: sending %d bytes in %p\n", GST_BUFFER_SIZE(buf), GST_BUFFER_DATA(buf));
    // copy the buffer
    memcpy(GST_BUFFER_DATA(buf), grab_buf, GST_BUFFER_SIZE(buf));

    if (v4lsrc->converter) {
      GST_DEBUG (0,"v4lsrc: converting\n");
      gst_colorspace_convert(v4lsrc->converter, GST_BUFFER_DATA(buf), GST_BUFFER_DATA(buf));
    }
    GST_DEBUG (0,"v4lsrc: sent %d bytes in %p\n", GST_BUFFER_SIZE(buf), GST_BUFFER_DATA(buf));

    gst_pad_push(v4lsrc->srcpad,buf);
  }
  else {
    gst_buffer_unref(buf);
  }
}

static void gst_v4lsrc_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstV4lSrc *src;
  int ret = 0;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_V4LSRC(object));
  src = GST_V4LSRC(object);

  switch (id) {
    case ARG_WIDTH:
      src->width = GTK_VALUE_INT(*arg);
      gst_v4lsrc_sync_parms(src);
      break;
    case ARG_HEIGHT:
      src->height = GTK_VALUE_INT(*arg);
      gst_v4lsrc_sync_parms(src);
      break;
    case ARG_FORMAT:
      src->format = GTK_VALUE_INT(*arg);
      break;
    case ARG_TUNE:
      ret = src->grabber->grab_tune(src->grabber, GTK_VALUE_ULONG(*arg));
      break;
    case ARG_INPUT:
      ret = src->grabber->grab_input(src->grabber, GTK_VALUE_INT(*arg), -1);
      break;
    case ARG_NORM:
      ret = src->grabber->grab_input(src->grabber, -1, GTK_VALUE_INT(*arg));
      break;
    case ARG_VOLUME:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_VOLUME, GTK_VALUE_INT(*arg));
      break;
    case ARG_MUTE:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_MUTE, GTK_VALUE_BOOL(*arg));
      break;
    case ARG_AUDIO_MODE:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_MODE, GTK_VALUE_INT(*arg));
      break;
    case ARG_COLOR:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_COLOR, GTK_VALUE_INT(*arg));
      break;
    case ARG_BRIGHT:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_BRIGHT, GTK_VALUE_INT(*arg));
      break;
    case ARG_HUE:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_HUE, GTK_VALUE_INT(*arg));
      break;
    case ARG_CONTRAST:
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_CONTRAST, GTK_VALUE_INT(*arg));
      break;
    default:
      ret = -1;
      break;
  }
  if (ret == -1) {
    fprintf(stderr, "v4lsrc: error setting property\n");
  }
}

static void gst_v4lsrc_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstV4lSrc *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_V4LSRC(object));
  src = GST_V4LSRC(object);

  switch (id) {
    case ARG_WIDTH:
      GTK_VALUE_INT(*arg) = src->width;
      break;
    case ARG_HEIGHT:
      GTK_VALUE_INT(*arg) = src->height;
      break;
    case ARG_TUNED:
      GTK_VALUE_BOOL(*arg) = src->grabber->grab_tuned(src->grabber);
      break;
    case ARG_VOLUME:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_VOLUME);
      break;
    case ARG_MUTE:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_MUTE);
      break;
    case ARG_AUDIO_MODE:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_MODE);
      break;
    case ARG_COLOR:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_COLOR);
      break;
    case ARG_BRIGHT:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_BRIGHT);
      break;
    case ARG_HUE:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_HUE);
      break;
    case ARG_CONTRAST:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_CONTRAST);
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static GstElementStateReturn gst_v4lsrc_change_state(GstElement *element) {
  g_return_val_if_fail(GST_IS_V4LSRC(element), FALSE);

  /* if going down into NULL state, close the file if it's open */
  if (GST_STATE_PENDING(element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET(element,GST_V4LSRC_OPEN))
      gst_v4lsrc_close_v4l(GST_V4LSRC(element));
  /* otherwise (READY or higher) we need to open the sound card */
  } else {
    if (!GST_FLAG_IS_SET(element,GST_V4LSRC_OPEN)) {
      if (!gst_v4lsrc_open_v4l(GST_V4LSRC(element)))
        return GST_STATE_FAILURE;
    }
  }

  if (GST_ELEMENT_CLASS(parent_class)->change_state)
    return GST_ELEMENT_CLASS(parent_class)->change_state(element);
  return GST_STATE_SUCCESS;
}

void gst_v4lsrc_sync_overlay(GtkWidget *src, MetaOverlay *raw, GstV4lSrc *v4lsrc) {

  if (raw->fully_obscured) {
    v4lsrc->grabber->grab_overlay(v4lsrc->grabber, 0, 0, 0, 0, 0, 0, 0);
    return;
  }

  v4lsrc->grabber->grab_overlay(v4lsrc->grabber, raw->wx, raw->wy,
		raw->width, raw->height, VIDEO_RGB16_LE, raw->overlay_clip, 
		raw->clip_count);

  v4lsrc->ov_wx = raw->wx;
  v4lsrc->ov_wy = raw->wy;
  v4lsrc->width = v4lsrc->ov_width = raw->width;
  v4lsrc->height = v4lsrc->ov_height = raw->height;
}

void gst_v4lsrc_sync_parms(GstV4lSrc *src) {
  int linelength;
  int format;
  int bytes;
  g_return_if_fail(src != NULL);
  g_return_if_fail(GST_IS_V4LSRC(src));

  GST_DEBUG (0,"v4lsrc: resync %d %d %d\n", src->width, src->height, src->format);

  src->converter = NULL;

  switch (src->format) {
    case GST_COLORSPACE_BGR555:
      format = VIDEO_RGB15_LE;
      bytes = 2;
      break;
    case GST_COLORSPACE_BGR565:
      format = VIDEO_RGB16_LE;
      bytes = 2;
      break;
    case GST_COLORSPACE_BGR24:
      format = VIDEO_BGR24;
      bytes = 3;
      break;
    case GST_COLORSPACE_RGB24:
      GST_DEBUG (0,"v4lsrc: initializing converter\n");
      format = VIDEO_BGR24;
      bytes = 3;
      if (src->converter) gst_colorspace_destroy(src->converter);
      src->converter = gst_colorspace_converter_new(src->width, src->height, GST_COLORSPACE_BGR24, GST_COLORSPACE_RGB24, NULL);
      break;
    case GST_COLORSPACE_BGR32:
      format = VIDEO_BGR32;
      bytes = 4;
      break;
    case GST_COLORSPACE_RGB32:
      GST_DEBUG (0,"v4lsrc: initializing converter\n");
      format = VIDEO_RGB32;
      bytes = 4;
      if (src->converter) gst_colorspace_destroy(src->converter);
      src->converter = gst_colorspace_converter_new(src->width, src->height, GST_COLORSPACE_BGR32, GST_COLORSPACE_RGB32, NULL);
      break;
    case GST_COLORSPACE_YUV420P:
      format = VIDEO_YUV420P;
      bytes = 2;
      break;
    case GST_COLORSPACE_YUV422P:
      format = VIDEO_YUV422P;
      bytes = 2;
      break;
    default:
      printf("v4lsrc: unsupported videoformat using RGB24\n");
      format = VIDEO_BGR24;
      src->format = GST_COLORSPACE_RGB24;
      bytes = 3;
      if (src->converter) gst_colorspace_destroy(src->converter);
      src->converter = gst_colorspace_converter_new(src->width, src->height, GST_COLORSPACE_BGR24, GST_COLORSPACE_RGB24, NULL);
      break;
  }

  src->buffer_size = src->width * src->height * bytes;
  if (!src->grabber->opened) return;

  if (src->grabber->grab_setparams(src->grabber, format, &src->width, &src->height, &linelength) != 0) {
    fprintf(stderr, "v4lsrc: error setting params\n");
  }
  GST_DEBUG (0,"v4lsrc: resynced to %d %d %d\n", src->width, src->height, src->buffer_size);

}

static gboolean gst_v4lsrc_open_v4l(GstV4lSrc *src) {
  g_return_val_if_fail(src->grabber != NULL, FALSE);
  g_return_val_if_fail(!src->grabber->opened, FALSE);

  if (src->grabber->grab_open(src->grabber, NULL) != -1) {
    gst_v4lsrc_sync_parms(src);
    GST_FLAG_SET(src, GST_V4LSRC_OPEN);
    return TRUE;
  }
  return FALSE;
}

static void gst_v4lsrc_close_v4l(GstV4lSrc *src) {
  g_return_if_fail(src->grabber != NULL);
  g_return_if_fail(src->grabber->opened);

  src->grabber->grab_close(src->grabber);
  GST_FLAG_UNSET(src, GST_V4LSRC_OPEN);
}

GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* this filter needs the riff parser */
  if (!gst_library_load("gstcolorspace")) {
    gst_info("v4lsrc: could not load support library: 'gstcolorspace'\n");
    return NULL;
  }

  plugin = gst_plugin_new("v4lsrc");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the v4lsrcparse element */
  factory = gst_elementfactory_new("v4lsrc",GST_TYPE_V4LSRC,
                                   &gst_v4lsrc_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}

