/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_V4LSRC_H__
#define __GST_V4LSRC_H__


#include <config.h>
#include <gst/gst.h>
#include <gst/meta/videoraw.h>
#include <libs/colorspace/gstcolorspace.h>

#include <linux/videodev.h>

#include "grab.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


GstElementDetails gst_v4lsrc_details;


#define GST_TYPE_V4LSRC \
  (gst_v4lsrc_get_type())
#define GST_V4LSRC(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_V4LSRC,GstV4lSrc))
#define GST_V4LSRC_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_V4LSRC,GstV4lSrcClass))
#define GST_IS_V4LSRC(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_V4LSRC))
#define GST_IS_V4LSRC_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_V4LSRC))

// NOTE: per-element flags start with 16 for now
typedef enum {
  GST_V4LSRC_OPEN            = GST_ELEMENT_FLAG_LAST,

  GST_V4LSRC_FLAG_LAST       = GST_ELEMENT_FLAG_LAST+2,
} GstV4lSrcFlags;

typedef struct _GstV4lSrc GstV4lSrc;
typedef struct _GstV4lSrcClass GstV4lSrcClass;

struct _GstV4lSrc {
  GstElement element;

  /* pads */
  GstPad *srcpad;

  /* video device */
  struct GRABBER *grabber;

  /* colorspace conversion */
  GstColorSpaceConverter *converter;
  
  // the current frame
  gint width;
  gint height;
  guint16 format;
  guint32 buffer_size;

  gboolean new_meta;
  gboolean overlay_ok;
  gboolean overlay_disable;
  gboolean in_overlay;
  gint ov_wx;
  gint ov_wy;
  gint ov_width;
  gint ov_height;

  GstMeta *meta;
};

struct _GstV4lSrcClass {
  GstElementClass parent_class;
};

GtkType gst_v4lsrc_get_type(void);
GstElement *gst_v4lsrc_new(gchar *name);

void gst_v4lsrc_push(GstElement *element);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_V4LSRC_H__ */
