/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

//#define GST_DEBUG_ENABLED

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#include <sys/soundcard.h>

#include "cdparanoia.h"


GstElementDetails cdparanoia_details = {
  "CD Audio (cdda) Source, Paranoia IV",
  "Source/File/CDDA",
  "Read audio from CD in paranoid mode",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 2000",
};

static GstPadFactory cdparanoia_src_factory = {
  "src",
  GST_PAD_FACTORY_SRC,
  GST_PAD_FACTORY_ALWAYS,
  GST_PAD_FACTORY_CAPS(
  "cdparanoia_src",
    "audio/raw",
    "format",   GST_PROPS_INT (AFMT_S16_LE),
    "depth",    GST_PROPS_INT (16),
    "rate",     GST_PROPS_INT (44100),
    "channels", GST_PROPS_INT (2),
    "chunksize", GST_PROPS_INT (CD_FRAMESIZE_RAW)
  ),
  NULL
};


/********** Define useful types for non-programmatic interfaces **********/
#define GST_TYPE_PARANOIA_MODE (gst_lame_paranoia_get_type())
static GtkType 
gst_lame_paranoia_get_type(void) {
  static GtkType paranoia_mode_type = 0;
  static GtkEnumValue paranoia_modes[] = {
    { 0, "0", "Disable paranoid checking" },
    { 1, "1", "cdda2wav-style overlap checking" },
    { 2, "2", "Full paranoia" },
    { 0, NULL, NULL },
  };
  if (!paranoia_mode_type) {
    paranoia_mode_type = gtk_type_register_enum ("GstParanoiaMode", paranoia_modes);
  }
  return paranoia_mode_type;
}


/********** Standard stuff for signals and arguments **********/
/* CDParanoia signals and args */
enum {
  SMILIE_CHANGE,
  TRANSPORT_ERROR,
  UNCORRECTED_ERROR,
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_LOCATION,
  ARG_GENERIC_DEVICE,
  ARG_START_TRACK,
  ARG_END_TRACK,
  ARG_START_SECTOR,
  ARG_END_SECTOR,
  ARG_DEFAULT_SECTORS,
  ARG_SEARCH_OVERLAP,
  ARG_ENDIAN,
  ARG_READ_SPEED,
  ARG_TOC_OFFSET,
  ARG_TOC_BIAS,
  ARG_NEVER_SKIP,
  ARG_ABORT_ON_SKIP,
  ARG_PARANOIA_MODE,
  ARG_SMILIE,
};


static void 			cdparanoia_class_init		(CDParanoiaClass *klass);
static void 			cdparanoia_init			(CDParanoia *cdparanoia);

static void 			cdparanoia_set_arg		(GtkObject *object,GtkArg *arg,guint id);
static void 			cdparanoia_get_arg		(GtkObject *object,GtkArg *arg,guint id);

static GstBuffer*		cdparanoia_get			(GstPad *pad);

static GstElementStateReturn 	cdparanoia_change_state 	(GstElement *element);


static GstPadTemplate *src_template;

static GstElementClass *parent_class = NULL;
static guint cdparanoia_signals[LAST_SIGNAL] = { 0 };

GtkType
cdparanoia_get_type (void) 
{
  static GtkType cdparanoia_type = 0;

  if (!cdparanoia_type) {
    static const GtkTypeInfo cdparanoia_info = {
      "CDParanoia",
      sizeof(CDParanoia),
      sizeof(CDParanoiaClass),
      (GtkClassInitFunc)cdparanoia_class_init,
      (GtkObjectInitFunc)cdparanoia_init,
      (GtkArgSetFunc)cdparanoia_set_arg,
      (GtkArgGetFunc)cdparanoia_get_arg,
      (GtkClassInitFunc)NULL,
    };
    cdparanoia_type = gtk_type_unique (GST_TYPE_ELEMENT, &cdparanoia_info);
  }
  return cdparanoia_type;
}

static void
cdparanoia_class_init (CDParanoiaClass *klass) 
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class (GST_TYPE_ELEMENT);

  cdparanoia_signals[SMILIE_CHANGE] =
    gtk_signal_new ("smilie_change", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (CDParanoiaClass, smilie_change),
                    gtk_marshal_NONE__STRING, GTK_TYPE_NONE, 1, GTK_TYPE_STRING);
  cdparanoia_signals[TRANSPORT_ERROR] =
    gtk_signal_new ("transport_error", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (CDParanoiaClass, transport_error),
                    gtk_marshal_NONE__INT, GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  cdparanoia_signals[UNCORRECTED_ERROR] =
    gtk_signal_new ("uncorrected_error", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (CDParanoiaClass, uncorrected_error),
                    gtk_marshal_NONE__INT, GTK_TYPE_NONE, 1, GTK_TYPE_INT);

  gtk_object_add_arg_type ("CDParanoia::location", GTK_TYPE_STRING,
                           GTK_ARG_READWRITE, ARG_LOCATION);
  gtk_object_add_arg_type ("CDParanoia::generic_device", GTK_TYPE_STRING,
                           GTK_ARG_READWRITE, ARG_GENERIC_DEVICE);
  gtk_object_add_arg_type ("CDParanoia::start_track", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_START_TRACK);
  gtk_object_add_arg_type ("CDParanoia::end_track", GTK_TYPE_INT,
                           GTK_ARG_READABLE, ARG_END_TRACK);
  gtk_object_add_arg_type ("CDParanoia::start_sector", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_START_SECTOR);
  gtk_object_add_arg_type ("CDParanoia::end_sector", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_END_SECTOR);
  gtk_object_add_arg_type ("CDParanoia::default_sectors", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_DEFAULT_SECTORS);
  gtk_object_add_arg_type ("CDParanoia::search_overlap", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_SEARCH_OVERLAP);
  gtk_object_add_arg_type ("CDParanoia::endian", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_ENDIAN);
  gtk_object_add_arg_type ("CDParanoia::read_speed", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_READ_SPEED);
  gtk_object_add_arg_type ("CDParanoia::toc_offset", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_TOC_OFFSET);
  gtk_object_add_arg_type ("CDParanoia::toc_bias", GTK_TYPE_BOOL,
                           GTK_ARG_READWRITE, ARG_TOC_BIAS);
  gtk_object_add_arg_type ("CDParanoia::never_skip", GTK_TYPE_BOOL,
                           GTK_ARG_READWRITE, ARG_NEVER_SKIP);
  gtk_object_add_arg_type ("CDParanoia::abort_on_skip", GTK_TYPE_BOOL,
                           GTK_ARG_READWRITE, ARG_ABORT_ON_SKIP);
  gtk_object_add_arg_type ("CDParanoia::paranoia_mode", GST_TYPE_PARANOIA_MODE,
                           GTK_ARG_READWRITE, ARG_PARANOIA_MODE);

  gtkobject_class->set_arg = cdparanoia_set_arg;
  gtkobject_class->get_arg = cdparanoia_get_arg;

  gstelement_class->change_state = cdparanoia_change_state;
}

static void 
cdparanoia_init (CDParanoia *cdparanoia) 
{
  cdparanoia->srcpad = gst_pad_new_from_template (src_template, "src");
  gst_pad_set_get_function (cdparanoia->srcpad, cdparanoia_get);
  gst_element_add_pad (GST_ELEMENT (cdparanoia), cdparanoia->srcpad);

  cdparanoia->device = "/dev/cdrom";
  cdparanoia->generic_device = NULL;
  cdparanoia->start_sector = -1;
  cdparanoia->end_sector = -1;
  cdparanoia->start_track = -1;
  cdparanoia->end_track = -1;
  cdparanoia->default_sectors = -1;
  cdparanoia->search_overlap = -1;
  cdparanoia->endian = -1;
  cdparanoia->read_speed = -1;
  cdparanoia->toc_offset = 0;
  cdparanoia->toc_bias = FALSE;
  cdparanoia->never_skip = FALSE;
  cdparanoia->paranoia_mode = 2;
  cdparanoia->abort_on_skip = FALSE;

  cdparanoia->cur_sector = 0;
  cdparanoia->seq = 0;
}


static void 
cdparanoia_set_arg (GtkObject *object, GtkArg *arg, guint id) 
{
  CDParanoia *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_CDPARANOIA (object));
  
  src = CDPARANOIA (object);

  switch(id) {
    case ARG_LOCATION:
      /* the element must be stopped in order to do this */
//      g_return_if_fail(!GST_FLAG_IS_SET(src,GST_STATE_RUNNING));

      if (src->device) g_free (src->device);
      /* clear the filename if we get a NULL (is that possible?) */
      if (GTK_VALUE_STRING (*arg) == NULL)
        src->device = NULL;
      /* otherwise set the new filename */
      else
        src->device = g_strdup (GTK_VALUE_STRING (*arg));
      break;
    case ARG_GENERIC_DEVICE:
      /* the element must be stopped in order to do this */
//      g_return_if_fail(!GST_FLAG_IS_SET(src,GST_STATE_RUNNING));

      if (src->generic_device) g_free (src->generic_device);
      /* reset the device if we get a NULL (is that possible?) */
      if (GTK_VALUE_STRING (*arg) == NULL)
        src->generic_device = NULL;
      /* otherwise set the new filename */
      else
        src->generic_device = g_strdup (GTK_VALUE_STRING (*arg));  
      break;
    case ARG_START_SECTOR:
      src->start_sector = GTK_VALUE_INT (*arg);
      break;
    case ARG_END_SECTOR:
      src->end_sector = GTK_VALUE_INT (*arg);
      break;
    case ARG_START_TRACK:
      src->start_track = GTK_VALUE_INT (*arg);
      break;
    case ARG_END_TRACK:
      src->end_track = GTK_VALUE_INT (*arg);
      break;
    case ARG_DEFAULT_SECTORS:
      src->default_sectors = GTK_VALUE_INT (*arg);
      break;
    case ARG_SEARCH_OVERLAP:
      src->search_overlap = GTK_VALUE_INT (*arg);
      break;
    case ARG_ENDIAN:
      src->endian = GTK_VALUE_INT (*arg);
      break;
    case ARG_READ_SPEED:
      src->read_speed = GTK_VALUE_INT (*arg);
      break;
    case ARG_TOC_OFFSET:
      src->toc_offset = GTK_VALUE_INT (*arg);
      break;
    case ARG_TOC_BIAS:
      src->toc_bias = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_NEVER_SKIP:
      src->never_skip = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_ABORT_ON_SKIP:
      src->abort_on_skip = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_PARANOIA_MODE:
      src->paranoia_mode = GTK_VALUE_INT (*arg);
      break;
    default:
      break;
  }

}

static void 
cdparanoia_get_arg (GtkObject *object,GtkArg *arg,guint id) 
{
  CDParanoia *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_CDPARANOIA (object));
  
  src = CDPARANOIA (object);

  switch (id) {
    case ARG_LOCATION:
      GTK_VALUE_STRING (*arg) = src->device;
      break;
    case ARG_GENERIC_DEVICE:
      GTK_VALUE_STRING (*arg) = src->generic_device;
      break;
    case ARG_START_SECTOR:
      GTK_VALUE_INT (*arg) = src->start_sector;
      break;
    case ARG_END_SECTOR:
      GTK_VALUE_INT (*arg) = src->end_sector;
      break;
    case ARG_START_TRACK:
      GTK_VALUE_INT (*arg) = src->start_track;
      break;
    case ARG_END_TRACK:
      GTK_VALUE_INT (*arg) = src->end_track;
      break;
    case ARG_DEFAULT_SECTORS:
      GTK_VALUE_INT (*arg) = src->default_sectors;
      break;
    case ARG_SEARCH_OVERLAP:
      GTK_VALUE_INT (*arg) = src->search_overlap;
      break;
    case ARG_ENDIAN:
      GTK_VALUE_INT (*arg) = src->endian;
      break;
    case ARG_READ_SPEED:
      GTK_VALUE_INT (*arg) = src->read_speed;
      break;
    case ARG_TOC_OFFSET:
      GTK_VALUE_INT (*arg) = src->toc_offset;
      break;
    case ARG_TOC_BIAS:
      GTK_VALUE_BOOL (*arg) = src->toc_bias;
      break;
    case ARG_NEVER_SKIP:
      GTK_VALUE_BOOL (*arg) = src->never_skip;
      break;
    case ARG_ABORT_ON_SKIP:
      GTK_VALUE_BOOL (*arg) = src->abort_on_skip;
      break;
    case ARG_PARANOIA_MODE:
      GTK_VALUE_INT (*arg) = src->paranoia_mode;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static void cdparanoia_callback(long inpos, int function) {
}

static GstBuffer* 
cdparanoia_get (GstPad *pad) 
{
  CDParanoia *src;
  GstBuffer *buf;
  static gint16 *cdda_buf;

  g_return_val_if_fail (pad != NULL, NULL);
  src = CDPARANOIA (gst_pad_get_parent (pad));
  g_return_val_if_fail (GST_FLAG_IS_SET (src, CDPARANOIA_OPEN), NULL);

  // create a new buffer
  buf = gst_buffer_new ();
  g_return_val_if_fail (buf, NULL);

  // read a sector
// NOTE: if you have a patched cdparanoia, set this to 1 to save cycles
#if 0
  cdda_buf = paranoia_read (src->p, NULL);
#else
  cdda_buf = paranoia_read (src->p, cdparanoia_callback);
#endif

  // update current sector and stop things if appropriate
  src->cur_sector++;
  if (src->cur_sector == src->end_sector) {
    GST_DEBUG (0,"setting EOS flag on outgoing buffer\n");
    GST_BUFFER_FLAG_SET (buf,GST_BUFFER_EOS);
  }

  // have to copy the buffer for now since we don't own it...
  // FIXME must ask monty about allowing ownership transfer
  GST_BUFFER_DATA (buf) = g_malloc(CD_FRAMESIZE_RAW);
  memcpy (GST_BUFFER_DATA (buf), cdda_buf, CD_FRAMESIZE_RAW);
  GST_BUFFER_SIZE (buf) = CD_FRAMESIZE_RAW;

  /* we're done, push the buffer off now */
  return buf;
}

/* open the file, necessary to go to RUNNING state */
static gboolean 
cdparanoia_open (CDParanoia *src) 
{
  gint i;
  gint paranoia_mode;

  g_return_val_if_fail (!GST_FLAG_IS_SET (src, CDPARANOIA_OPEN), FALSE);

  GST_DEBUG_ENTER("(\"%s\",...)",gst_element_get_name(GST_ELEMENT(src)));

  // find the device
  if (src->generic_device != NULL) {
    src->d = cdda_identify_scsi (src->generic_device, src->device, FALSE, NULL);
  } else {
    if (src->device != NULL) {
      src->d = cdda_identify (src->device, FALSE, NULL);
    } else {
      src->d = cdda_identify ("/dev/cdrom", FALSE, NULL);
    }
  }

  // fail if the device couldn't be found
  if (src->d == NULL) {
    GST_DEBUG (0,"couldn't open device\n");
    return FALSE;
  }

  // set verbosity mode
  cdda_verbose_set (src->d, CDDA_MESSAGE_FORGETIT, CDDA_MESSAGE_FORGETIT);

  // set various other parameters
  if (src->default_sectors != -1) {
    if ((src->default_sectors < 0) || (src->default_sectors > 100)) {
      GST_DEBUG (0,"default sector read size must be 1 <= n <= 100\n");
      cdda_close (src->d);
      src->d = NULL;
      return FALSE;
    } else {
      src->d->nsectors = src->default_sectors;
      src->d->bigbuff = src->default_sectors * CD_FRAMESIZE_RAW;
    }
  }
  if (src->search_overlap != -1) {
    if ((src->search_overlap < 0) || (src->search_overlap > 75)) {
      GST_DEBUG (0,"search overlap must be 0 <= n <= 75\n");
      cdda_close (src->d);
      src->d = NULL;
      return FALSE;
    }
  }

  // open the disc
  if (cdda_open (src->d)) {
    GST_DEBUG (0,"couldn't open disc\n");
    cdda_close (src->d);
    src->d = NULL;
    return FALSE;
  }

  // set up some more stuff
  if (src->toc_bias) {
    src->toc_offset -= cdda_track_firstsector (src->d, 1);
  }
  for (i=0;i<src->d->tracks + 1;i++) {
    src->d->disc_toc[i].dwStartSector += src->toc_offset;
  }

  if (src->read_speed != -1) {
    cdda_speed_set (src->d, src->read_speed);
  }

  // if the start_track is set, override the start_sector
  if (src->start_track != -1) {
    src->start_sector = cdda_track_firstsector (src->d, src->start_track);
  // if neither start_track nor start_sector is set, 
  } else if (src->start_sector == -1) {
    src->start_sector = cdda_disc_firstsector (src->d);
  }
  // if the end_track is set, override the end_sector
  if (src->end_track != -1) {
    src->end_sector = cdda_track_lastsector (src->d, src->end_track);
  // if neither end_track nor end_sector is set, 
  } else if (src->end_sector == -1) {
    src->end_sector = cdda_disc_lastsector (src->d);
  }

  // create the paranoia struct and set it up
  src->p = paranoia_init (src->d);
  if (src->p == NULL) {
    GST_DEBUG (0,"couldn't create paranoia struct\n");
    return FALSE;
  }

  if (src->paranoia_mode == 0) paranoia_mode = PARANOIA_MODE_DISABLE;
  else if (src->paranoia_mode == 1) paranoia_mode = PARANOIA_MODE_OVERLAP;
  else paranoia_mode = PARANOIA_MODE_FULL;
  if (src->never_skip) paranoia_mode |= PARANOIA_MODE_NEVERSKIP;
  paranoia_modeset (src->p, paranoia_mode);

  if (src->search_overlap != -1) {
    paranoia_overlapset (src->p, src->search_overlap);
  }

  src->cur_sector = src->start_sector;
  paranoia_seek (src->p, src->cur_sector, SEEK_SET);
  GST_DEBUG (0,"successfully seek'd to beginning of disk\n");

  GST_FLAG_SET (src, CDPARANOIA_OPEN);

  GST_DEBUG_LEAVE("");
  
  return TRUE;
}

/* close the file */
static void 
cdparanoia_close (CDParanoia *src)
{
  g_return_if_fail (GST_FLAG_IS_SET (src, CDPARANOIA_OPEN));

  // kill the paranoia state
  paranoia_free (src->p);
  src->p = NULL;

  // close the disc
  cdda_close (src->d);
  src->d = NULL;

  GST_FLAG_UNSET (src, CDPARANOIA_OPEN);
}

static GstElementStateReturn
cdparanoia_change_state (GstElement *element)
{
  g_return_val_if_fail (GST_IS_CDPARANOIA (element), GST_STATE_FAILURE);

  g_print("state pending %d\n", GST_STATE_PENDING (element));

  /* if going down into NULL state, close the file if it's open */
  if (GST_STATE_PENDING (element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET (element, CDPARANOIA_OPEN))
      cdparanoia_close (CDPARANOIA (element));
  /* otherwise (READY or higher) we need to open the file */
  } else {
    if (!GST_FLAG_IS_SET (element, CDPARANOIA_OPEN)) {
      if (!cdparanoia_open (CDPARANOIA (element))) {
        g_print("failed opening cd\n");
        return GST_STATE_FAILURE;
      }
    }
  }

  /* if we haven't failed already, give the parent class a chance too ;-) */
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

GstPlugin *
plugin_init (GModule *module) 
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new ("cdparanoia");
  g_return_val_if_fail (plugin != NULL, NULL);

  /* create an elementfactory for the cdparanoia element */
  factory = gst_elementfactory_new ("cdparanoia", GST_TYPE_CDPARANOIA,
                                    &cdparanoia_details);
  g_return_val_if_fail (factory != NULL, NULL);

  /* register the source's caps */
  src_template = gst_padtemplate_new (&cdparanoia_src_factory);
  gst_elementfactory_add_padtemplate (factory, src_template);

  /* and add the cdparanoia element factory to the plugin */
  gst_plugin_add_factory (plugin, factory);
  
  return plugin;
}
