/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stereo.h>


GstElementDetails stereo_details = {
  "Stereo effect",
  "Filter/Effect",
  "Muck with the stereo signal, enhance it's 'stereo-ness'",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* Stereo signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_ACTIVE,
  ARG_STEREO
};


static void gst_stereo_class_init(GstStereoClass *klass);
static void gst_stereo_init(GstStereo *stereo);
static void gst_stereo_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_stereo_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstElementClass *parent_class = NULL;
//static guint gst_stereo_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_stereo_get_type(void) {
  static GtkType stereo_type = 0;

  if (!stereo_type) {
    static const GtkTypeInfo stereo_info = {
      "GstStereo",
      sizeof(GstStereo),
      sizeof(GstStereoClass),
      (GtkClassInitFunc)gst_stereo_class_init,
      (GtkObjectInitFunc)gst_stereo_init,
      (GtkArgSetFunc)gst_stereo_set_arg,
      (GtkArgGetFunc)gst_stereo_get_arg,
      (GtkClassInitFunc)NULL,
    };
    stereo_type = gtk_type_unique(GST_TYPE_ELEMENT,&stereo_info);
  }
  return stereo_type;
}

static void gst_stereo_class_init(GstStereoClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstStereo::active", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_ACTIVE);
  gtk_object_add_arg_type("GstStereo::stereo", GTK_TYPE_FLOAT,
                          GTK_ARG_READWRITE, ARG_STEREO);

  gtkobject_class->set_arg = gst_stereo_set_arg;  
  gtkobject_class->get_arg = gst_stereo_get_arg;
}

static void gst_stereo_init(GstStereo *stereo) {
  stereo->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(stereo),stereo->sinkpad);
  gst_pad_set_chain_function(stereo->sinkpad,gst_stereo_chain);
  stereo->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(stereo),stereo->srcpad);

  stereo->active = FALSE;
  stereo->stereo = 2.5;
}

GstElement *gst_stereo_new(gchar *name) {
  GstElement *stereo = GST_ELEMENT(gtk_type_new(GST_TYPE_STEREO));
  gst_element_set_name(GST_ELEMENT(stereo),name);
  return stereo;
}

void gst_stereo_chain(GstPad *pad,GstBuffer *buf) {
  GstStereo *stereo;
  gint16 *data;
  gint samples;
  gint i;
  gdouble avg,ldiff,rdiff,tmp,mul;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  stereo = GST_STEREO(pad->parent);
  g_return_if_fail(stereo != NULL);
  g_return_if_fail(GST_IS_STEREO(stereo));

//  FIXME
//  if (buf->meta)
//    memcpy(&stereo->meta,buf->meta,sizeof(stereo->meta));

  if (stereo->active) {

    if (stereo->meta.channels == 2 && stereo->meta.format == AFMT_S16_LE) {
      data = (gint16 *)GST_BUFFER_DATA(buf);
      samples = GST_BUFFER_SIZE(buf) / 2;
      mul = stereo->stereo;
      for (i = 0; i < samples / 2; i += 2) {
        avg = (data[i] + data[i + 1]) / 2;
        ldiff = data[i] - avg;
        rdiff = data[i + 1] - avg;

        tmp = avg + ldiff * mul;
        if (tmp < -32768)
          tmp = -32768;
        if (tmp > 32767)
          tmp = 32767;
        data[i] = tmp;

        tmp = avg + rdiff * mul;
        if (tmp < -32768)
          tmp = -32768;
        if (tmp > 32767)
          tmp = 32767;
        data[i + 1] = tmp;
      }
    }
  }

  gst_pad_push(stereo->srcpad,buf);
}

static void gst_stereo_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstStereo *stereo;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_STEREO(object));
  stereo = GST_STEREO(object);

  switch(id) {
    case ARG_ACTIVE:
      stereo->active = GTK_VALUE_INT(*arg);
      break;
    case ARG_STEREO:
      stereo->stereo = GTK_VALUE_FLOAT(*arg) * 10.0;
      break;
    default:
      break;
  }
}

static void gst_stereo_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstStereo *stereo;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_STEREO(object));
  stereo = GST_STEREO(object);

  switch (id) {
    case ARG_ACTIVE:
      GTK_VALUE_INT(*arg) = stereo->active;
      break;
    case ARG_STEREO:
      GTK_VALUE_FLOAT(*arg) = stereo->stereo / 10.0;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("stereo");
  g_return_val_if_fail(plugin != NULL, NULL);

  factory = gst_elementfactory_new("stereo",GST_TYPE_STEREO,
                                   &stereo_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
