/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <volume.h>

#ifdef HAVE_LIBMMX
#include <mmx.h>
#endif /* HAVE_LIBMMX */

GstElementDetails volume_details = {
  "Volume control",
  "Filter/Effect",
  "Digitally control the volume of a signal",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* Volume signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_ACTIVE,
  ARG_VOLUME
};


static void gst_volume_class_init(GstVolumeClass *klass);
static void gst_volume_init(GstVolume *volume);
static void gst_volume_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_volume_get_arg(GtkObject *object,GtkArg *arg,guint id);

/* various optimized volume control functions */
void gst_volume_chain_none(GstPad *pad,GstBuffer *buf);
void gst_volume_chain_fp_S16(GstPad *pad,GstBuffer *buf);
void gst_volume_chain_fp_S8(GstPad *pad,GstBuffer *buf);

void gst_volume_set_chain(GstVolume *volume,gint format);
void gst_volume_rechain(gint format,GstPad *pad,GstBuffer *buf);

static GstElementClass *parent_class = NULL;
//static guint gst_volume_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_volume_get_type(void) {
  static GtkType volume_type = 0;

  if (!volume_type) {
    static const GtkTypeInfo volume_info = {
      "GstVolume",
      sizeof(GstVolume),
      sizeof(GstVolumeClass),
      (GtkClassInitFunc)gst_volume_class_init,
      (GtkObjectInitFunc)gst_volume_init,
      (GtkArgSetFunc)gst_volume_set_arg,
      (GtkArgGetFunc)gst_volume_get_arg,
      (GtkClassInitFunc)NULL,
    };
    volume_type = gtk_type_unique(GST_TYPE_ELEMENT,&volume_info);
  }
  return volume_type;
}

static void gst_volume_class_init(GstVolumeClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstVolume::active", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_ACTIVE);
  gtk_object_add_arg_type("GstVolume::volume", GTK_TYPE_FLOAT,
                          GTK_ARG_READWRITE, ARG_VOLUME);

  gtkobject_class->set_arg = gst_volume_set_arg;  
  gtkobject_class->get_arg = gst_volume_get_arg;

#ifdef HAVE_LIBMMX
  //klass->mmx = mm_support();
#endif /* HAVE_LIBMMX */
}

static void gst_volume_init(GstVolume *volume) {
  volume->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(volume),volume->sinkpad);
  volume->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(volume),volume->srcpad);

  volume->active = FALSE;
  volume->volume = 1.0;

  /* decide on the chain function (well, it'll be _none()) */
  gst_volume_set_chain(volume,0);
}

GstElement *gst_volume_new(gchar *name) {
  GstElement *volume = GST_ELEMENT(gtk_type_new(GST_TYPE_VOLUME));
  gst_element_set_name(GST_ELEMENT(volume),name);
  return volume;
}

/* does no volume control whatsoever, just passes the buffer through */
void gst_volume_chain_none(GstPad *pad,GstBuffer *buf) {
  GstVolume *volume = GST_VOLUME(pad->parent);

  gst_pad_push(volume->srcpad,buf);
}

/* does arbitrary scaling of S16 data with floating point */
void gst_volume_chain_fp_S16(GstPad *pad,GstBuffer *buf) {
  GstVolume *volume = GST_VOLUME(pad->parent);
  gint16 *data;
  int i;

  /* check the format */
//  if (((MetaAudioRaw *)(buf->meta))->format != AFMT_S16_LE) {
//    gst_volume_rechain(((MetaAudioRaw *)(buf->meta))->format,pad,buf);
//    return;
//  }

  /* FIXME: we modify the buffer in place, no regard to CoW */
  data = (gint16 *)GST_BUFFER_DATA(buf);
  for (i=0;i<GST_BUFFER_SIZE(buf)/2;i++)
    data[i] = (float)data[i] * volume->volume;

  gst_pad_push(volume->srcpad,buf);
}

/* does arbitrary scaling of S8 data with floating point */
void gst_volume_chain_fp_S8(GstPad *pad,GstBuffer *buf) {
  GstVolume *volume = GST_VOLUME(pad->parent);
  gint8 *data;
  int i;

  /* check the format */
//  if (((MetaAudioRaw *)(buf->meta))->format != AFMT_S8) {
//    gst_volume_rechain(((MetaAudioRaw *)(buf->meta))->format,pad,buf);
//    return;
//  }

  /* FIXME: we modify the buffer in place, no regard to CoW */
  data = (gint8 *)GST_BUFFER_DATA(buf);
  for (i=0;i<GST_BUFFER_SIZE(buf)/2;i++)
    data[i] = (float)data[i] * volume->volume;

  gst_pad_push(volume->srcpad,buf);
}

/* set the chain function relative to the format, volume, etc. */
/* NOTE: this will change as I add more specialized chain functions */
/* FIXME: should add shifter functions for 2^ volume changes */
void gst_volume_set_chain(GstVolume *volume,gint format) {
  /* simple cases first */
  if ((volume->volume == 1.0) || !volume->active)
    gst_pad_set_chain_function(volume->sinkpad,gst_volume_chain_none);
  /* if we're lowering the volume... */
  else if (volume->volume < 1.0) {
    if (format == AFMT_S16_LE)
      gst_pad_set_chain_function(volume->sinkpad,gst_volume_chain_fp_S16);
    else if (format == AFMT_S8)
      gst_pad_set_chain_function(volume->sinkpad,gst_volume_chain_fp_S8);
  /* if we're raising the volume */
  } else if (volume->volume > 1.0) {
    if (format == AFMT_S16_LE)
      gst_pad_set_chain_function(volume->sinkpad,gst_volume_chain_fp_S16);
    else if (format == AFMT_S8)
      gst_pad_set_chain_function(volume->sinkpad,gst_volume_chain_fp_S8);
  }
}

void gst_volume_rechain(gint format,GstPad *pad,GstBuffer *buf) {
  gst_volume_set_chain(GST_VOLUME(pad->parent),format);
  pad->chainfunc(pad,buf);
}

static void gst_volume_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstVolume *volume;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VOLUME(object));
  volume = GST_VOLUME(object);

  switch(id) {
    case ARG_ACTIVE:
      volume->active = GTK_VALUE_INT(*arg);
      gst_volume_set_chain(volume,0);
      break;
    case ARG_VOLUME:
      volume->volume = GTK_VALUE_FLOAT(*arg);
      /* FIXME: this is cheesy, we set it to the most common, and if we're
         wrong it gets reset when the next buffer comes in.  maybe store
         the current format somewhere so we don't do this? */
      gst_volume_set_chain(volume,AFMT_S16_LE);
      break;
    default:
      break;
  }
}

static void gst_volume_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstVolume *volume;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VOLUME(object));
  volume = GST_VOLUME(object);

  switch (id) {
    case ARG_ACTIVE:
      GTK_VALUE_INT(*arg) = volume->active;
      break;
    case ARG_VOLUME:
      GTK_VALUE_FLOAT(*arg) = volume->volume;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("volume");
  g_return_val_if_fail(plugin != NULL, NULL);

  factory = gst_elementfactory_new("volume",GST_TYPE_VOLUME,
                                   &volume_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
