/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

//#define DEBUG_ENABLED
#include <smooth.h>


GstElementDetails smooth_details = {
  "Smooth effect",
  "Filter/Effect",
  "apply a smooth filter to an image",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};


/* Smooth signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_ACTIVE,
  ARG_TOLERANCE,
  ARG_FILTERSIZE
};

static void gst_smooth_class_init(GstSmoothClass *klass);
static void gst_smooth_init(GstSmooth *smooth);
static void smooth_filter(unsigned char* dest, unsigned char* src, int width, int height, int tolerance, int filtersize);

static void gst_smooth_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_smooth_get_arg(GtkObject *object,GtkArg *arg,guint id);

static GstElementClass *parent_class = NULL;
//static guint gst_smooth_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_smooth_get_type(void) {
  static GtkType smooth_type = 0;

  if (!smooth_type) {
    static const GtkTypeInfo smooth_info = {
      "GstSmooth",
      sizeof(GstSmooth),
      sizeof(GstSmoothClass),
      (GtkClassInitFunc)gst_smooth_class_init,
      (GtkObjectInitFunc)gst_smooth_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    smooth_type = gtk_type_unique(GST_TYPE_ELEMENT,&smooth_info);
  }
  return smooth_type;
}

static void gst_smooth_class_init(GstSmoothClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstSmooth::active", GTK_TYPE_BOOL,
		                             GTK_ARG_READWRITE, ARG_ACTIVE);
  gtk_object_add_arg_type("GstSmooth::tolerance", GTK_TYPE_INT,
		                             GTK_ARG_READWRITE, ARG_TOLERANCE);
  gtk_object_add_arg_type("GstSmooth::filtersize", GTK_TYPE_INT,
		                               GTK_ARG_READWRITE, ARG_FILTERSIZE);

  gtkobject_class->set_arg = gst_smooth_set_arg;
  gtkobject_class->get_arg = gst_smooth_get_arg;

}

static void gst_smooth_init(GstSmooth *smooth) {
  smooth->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(smooth),smooth->sinkpad);
  gst_pad_set_chain_function(smooth->sinkpad,gst_smooth_chain);
  smooth->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(smooth),smooth->srcpad);

  smooth->active = TRUE;
  smooth->tolerance = 8;
  smooth->filtersize = 3;
}

GstElement *gst_smooth_new(gchar *name) {
  GstElement *smooth = GST_ELEMENT(gtk_type_new(GST_TYPE_SMOOTH));
  gst_element_set_name(GST_ELEMENT(smooth),name);
  return smooth;
}

static void smooth_filter(unsigned char* dest, unsigned char* src, int width, int height, int tolerance, int filtersize)
{
  int refval, aktval, upperval, lowerval, numvalues, sum;
  int x, y, fx, fy, fy1, fy2, fx1, fx2;
  unsigned char *srcp = src;

  fy1 = 0;
  fy2 = MIN(filtersize+1, height) * width;
		    
  for(y = 0; y < height; y++)
  {
    if (y>(filtersize+1)) fy1 += width;
    if (y<height-(filtersize+1)) fy2 += width;

    for(x = 0; x < width; x++)
    {
      refval    = *src;
      upperval  = refval + tolerance;
      lowerval  = refval - tolerance;
				
      numvalues = 1;
      sum       = refval;

      fx1      = MAX(x-filtersize,   0)     + fy1;
      fx2      = MIN(x+filtersize+1, width) + fy1;

      for (fy = fy1; fy<fy2; fy+=width)
      {
        for (fx = fx1; fx<fx2; fx++)
        {
          aktval = srcp[fx];
          if ((lowerval-aktval)*(upperval-aktval)<0)
          {
            numvalues ++;
            sum += aktval;
          }
        } //for fx
        fx1 += width;
        fx2 += width;
      } //for fy

      src++;
      *dest++ = sum/numvalues;
    }
  }
}

void gst_smooth_chain(GstPad *pad,GstBuffer *buf) {
  GstSmooth *smooth;
  guchar *data;
  gulong size;
  GstBuffer *outbuf;
  GstMeta *meta;
  int lumsize, chromsize;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  smooth = GST_SMOOTH(pad->parent);

  if (!smooth->active) {
    gst_pad_push(smooth->srcpad,buf);
    return;
  }

  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"smooth: have buffer of %d\n", GST_BUFFER_SIZE(buf));
  
  outbuf = gst_buffer_new();
  GST_BUFFER_DATA(outbuf) = g_malloc(GST_BUFFER_SIZE(buf));
  GST_BUFFER_SIZE(outbuf) = GST_BUFFER_SIZE(buf);

  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (smooth->format != ((MetaVideoRaw *)meta)->format ||
        smooth->width  != ((MetaVideoRaw *)meta)->width ||
        smooth->height != ((MetaVideoRaw *)meta)->height) {

      smooth->format = ((MetaVideoRaw *)meta)->format;
      smooth->width  = ((MetaVideoRaw *)meta)->width;
      smooth->height = ((MetaVideoRaw *)meta)->height;

      gst_buffer_add_meta(outbuf, meta);
    }
  }
  lumsize = smooth->width*smooth->height;
  chromsize = lumsize/4;

  smooth_filter(GST_BUFFER_DATA(outbuf), data, smooth->width, smooth->height, smooth->tolerance, smooth->filtersize);
  smooth_filter(GST_BUFFER_DATA(outbuf)+lumsize, data+lumsize, smooth->width/2, smooth->height/2, smooth->tolerance, smooth->filtersize/2);
  smooth_filter(GST_BUFFER_DATA(outbuf)+lumsize+chromsize, data+lumsize+chromsize, smooth->width/2, smooth->height/2, smooth->tolerance, smooth->filtersize/2);

  gst_buffer_unref(buf);

  gst_pad_push(smooth->srcpad,outbuf);

}

static void gst_smooth_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstSmooth *smooth;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_SMOOTH(object));
  smooth = GST_SMOOTH(object);

  switch(id) {
    case ARG_ACTIVE:
      smooth->active = GTK_VALUE_BOOL(*arg);
      break;
    case ARG_TOLERANCE:
      smooth->tolerance = GTK_VALUE_INT(*arg);
      break;
    case ARG_FILTERSIZE:
      smooth->filtersize = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void gst_smooth_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstSmooth *smooth;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_SMOOTH(object));
  smooth = GST_SMOOTH(object);

  switch (id) {
    case ARG_ACTIVE:
      GTK_VALUE_BOOL(*arg) = smooth->active;
      break;
    case ARG_TOLERANCE:
      GTK_VALUE_INT(*arg) = smooth->tolerance;
      break;
    case ARG_FILTERSIZE:
      GTK_VALUE_INT(*arg) = smooth->filtersize;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}


GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("smooth");
  g_return_val_if_fail(plugin != NULL, NULL);

  factory = gst_elementfactory_new("smooth",GST_TYPE_SMOOTH,
                                   &smooth_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
