/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_JPEGDEC_H__
#define __GST_JPEGDEC_H__


#include <config.h>
#include <gst/gst.h>
#include <gst/meta/videoraw.h>

#include <jpeglib.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GST_TYPE_JPEGDEC \
  (gst_jpegdec_get_type())
#define GST_JPEGDEC(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_JPEGDEC,GstJpegDec))
#define GST_JPEGDEC_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_JPEGDEC,GstJpegDec))
#define GST_IS_JPEGDEC(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_JPEGDEC))
#define GST_IS_JPEGDEC_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_JPEGDEC))

typedef struct _GstJpegDec GstJpegDec;
typedef struct _GstJpegDecClass GstJpegDecClass;

struct _GstJpegDec {
  GstElement element;

  /* pads */
  GstPad *sinkpad,*srcpad;

  int parse_state;
  // the timestamp of the next frame
  guint64 next_time;
  // the interval between frames
  guint64 time_interval;

  // video state
  gint format;
  gint width;
  gint height;
  // the size of the output buffer
  gint outsize;
  // the jpeg line buffer
  guchar **line[3];

  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  struct jpeg_source_mgr jsrc;
};

struct _GstJpegDecClass {
  GstElementClass parent_class;
};

GtkType gst_jpegdec_get_type(void);
GstElement *gst_jpegdec_new(gchar *name);
void gst_jpegdec_chain(GstPad *pad,GstBuffer *buf);


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_JPEGDEC_H__ */
