/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>

//#define DEBUG_ENABLED
#include "gstjpegenc.h"

/* elementfactory information */
GstElementDetails gst_jpegenc_details = {
  "jpeg image encoder",
  "Filter/Encoder/Image",
  ".jpeg",
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

/* JpegEnc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};

static void		gst_jpegenc_class_init	(GstJpegEnc *klass);
static void		gst_jpegenc_init	(GstJpegEnc *jpegenc);

GstElement		*gst_jpegenc_new	(gchar *name);
static void		gst_jpegenc_chain	(GstPad *pad,GstBuffer *buf);
static GstBuffer	*gst_jpegenc_get	(GstPad *pad);

static void		gst_jpegenc_resync	(GstJpegEnc *jpegenc); 

static GstElementClass *parent_class = NULL;
//static guint gst_jpegenc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_jpegenc_get_type(void) {
  static GtkType jpegenc_type = 0;

  if (!jpegenc_type) {
    static const GtkTypeInfo jpegenc_info = {
      "GstJpegEnc",
      sizeof(GstJpegEnc),
      sizeof(GstJpegEnc),
      (GtkClassInitFunc)gst_jpegenc_class_init,
      (GtkObjectInitFunc)gst_jpegenc_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    jpegenc_type = gtk_type_unique(GST_TYPE_ELEMENT,&jpegenc_info);
  }
  return jpegenc_type;
}

static void
gst_jpegenc_class_init(GstJpegEnc *klass) {
  GstElementClass *gstelement_class;

  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);
}

void gst_jpegenc_init_destination(j_compress_ptr cinfo) {
  GST_DEBUG (0,"gst_jpegenc_chain: init_destination\n");
}
gboolean gst_jpegenc_flush_destination(j_compress_ptr cinfo) {
  GST_DEBUG (0,"gst_jpegenc_chain: flush_destination: buffer too small !!!\n");
  return TRUE;
}
void gst_jpegenc_term_destination(j_compress_ptr cinfo) {
  GST_DEBUG (0,"gst_jpegenc_chain: term_source\n");
}

static void gst_jpegenc_init(GstJpegEnc *jpegenc) {
  /* create the sink and src pads */
  jpegenc->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(jpegenc),jpegenc->sinkpad);
  gst_pad_set_chain_function(jpegenc->sinkpad,gst_jpegenc_chain);
  gst_pad_set_get_function(jpegenc->sinkpad,gst_jpegenc_get);
  jpegenc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(jpegenc),jpegenc->srcpad);

  // reset the initial video state
  jpegenc->format = GST_COLORSPACE_YUV420P;
  jpegenc->width = -1;
  jpegenc->height = -1;

  /* setup jpeglib */
  memset(&jpegenc->cinfo, 0, sizeof(jpegenc->cinfo));
  memset(&jpegenc->jerr, 0, sizeof(jpegenc->jerr));
  jpegenc->cinfo.err = jpeg_std_error(&jpegenc->jerr);
  jpeg_create_compress(&jpegenc->cinfo);

  GST_DEBUG (0,"gst_jpegenc_init: setting line buffers\n");
  jpegenc->line[0] = NULL;
  jpegenc->line[1] = NULL;
  jpegenc->line[2] = NULL;

  gst_jpegenc_resync(jpegenc);

  jpegenc->jdest.init_destination = gst_jpegenc_init_destination;
  jpegenc->jdest.empty_output_buffer = gst_jpegenc_flush_destination;
  jpegenc->jdest.term_destination = gst_jpegenc_term_destination;
  jpegenc->cinfo.dest = &jpegenc->jdest;

}

GstElement *gst_jpegenc_new(gchar *name) {
  GstElement *jpegenc = GST_ELEMENT(gtk_type_new(GST_TYPE_JPEGENC));
  gst_element_set_name(GST_ELEMENT(jpegenc),name);
  return jpegenc;
}

static void gst_jpegenc_resync(GstJpegEnc *jpegenc) {
  guint size;
  gint width, height;

  GST_DEBUG (0,"gst_jpegenc_resync: resync\n");

  jpegenc->cinfo.image_width = width = jpegenc->width;
  jpegenc->cinfo.image_height = height = jpegenc->height;
  jpegenc->cinfo.input_components = 3;

  GST_DEBUG (0,"gst_jpegenc_resync: wdith %d, height %d\n", width, height);

  jpeg_set_defaults(&jpegenc->cinfo);
  jpegenc->cinfo.dct_method = JDCT_FASTEST;
  //jpegenc->cinfo.dct_method = JDCT_DEFAULT;
  //jpegenc->cinfo.smoothing_factor = 10;
  jpeg_set_quality(&jpegenc->cinfo, 85, TRUE);

  switch (jpegenc->format) {
    case GST_COLORSPACE_RGB24:
      size = 3;
      GST_DEBUG (0,"gst_jpegenc_resync: setting format to RGB24\n");
      jpegenc->cinfo.in_color_space = JCS_RGB;
      jpegenc->cinfo.raw_data_in = FALSE;
      break;
    case GST_COLORSPACE_YUV420P:
      size = 2;
      jpegenc->cinfo.raw_data_in = TRUE;
      jpegenc->cinfo.in_color_space = JCS_YCbCr;
      GST_DEBUG (0,"gst_jpegenc_resync: setting format to YUV420P\n");
      jpegenc->cinfo.comp_info[0].h_samp_factor = 2;
      jpegenc->cinfo.comp_info[0].v_samp_factor = 2;
      jpegenc->cinfo.comp_info[1].h_samp_factor = 1;
      jpegenc->cinfo.comp_info[1].v_samp_factor = 1;
      jpegenc->cinfo.comp_info[2].h_samp_factor = 1;
      jpegenc->cinfo.comp_info[2].v_samp_factor = 1;

      if (height != -1) {
        jpegenc->line[0] = g_realloc(jpegenc->line[0], height*sizeof(char*));
        jpegenc->line[1] = g_realloc(jpegenc->line[1], height*sizeof(char*)/2);
        jpegenc->line[2] = g_realloc(jpegenc->line[2], height*sizeof(char*)/2);
      }

      GST_DEBUG (0,"gst_jpegenc_resync: setting format done\n");
      break;
    default:
      printf("gst_jpegenc_resync: unsupported colorspace, using RGB\n");
      size = 3;
      jpegenc->cinfo.in_color_space = JCS_RGB;
      break;
  }
  jpegenc->bufsize = jpegenc->width*jpegenc->height*size;
  jpegenc->row_stride = width * size;

  jpeg_suppress_tables(&jpegenc->cinfo, TRUE);

  jpegenc->buffer = NULL;
  GST_DEBUG (0,"gst_jpegenc_resync: resync done\n");
}

static GstBuffer
*gst_jpegenc_get(GstPad *pad)
{
  GstJpegEnc *jpegenc;
  GstMeta *meta;
  GstBuffer *newbuf;

  GST_DEBUG (0,"gst_jpegenc_chain: pull buffer\n");

  g_return_val_if_fail(pad != NULL, NULL);
  g_return_val_if_fail(GST_IS_PAD(pad), NULL);

  jpegenc = GST_JPEGENC(pad->parent);

  if (jpegenc->buffer == NULL || GST_BUFFER_REFCOUNT(jpegenc->buffer) != 1) {
    if (jpegenc->buffer) gst_buffer_unref(jpegenc->buffer);
    GST_DEBUG (0,"gst_jpegenc_chain: new buffer\n");
    newbuf = jpegenc->buffer = gst_buffer_new();
    GST_BUFFER_DATA(newbuf) = g_malloc(jpegenc->bufsize);
    GST_BUFFER_SIZE(newbuf) = jpegenc->bufsize;

    meta = gst_meta_new_size(sizeof(MetaVideoRaw));

    ((MetaVideoRaw *)meta)->format = jpegenc->format;
    ((MetaVideoRaw *)meta)->width = jpegenc->width;
    ((MetaVideoRaw *)meta)->height = jpegenc->height;

    gst_buffer_add_meta(newbuf, meta);
  }
  gst_buffer_ref(jpegenc->buffer);

  return jpegenc->buffer;
}

static void gst_jpegenc_chain(GstPad *pad,GstBuffer *buf) {
  GstJpegEnc *jpegenc;
  guchar *data, *outdata;
  gulong size, outsize;
  GstBuffer *outbuf;
  GstMeta *meta;
  guint height, width, width2;
  guchar *base[3];
  gint i,j, k;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  //g_return_if_fail(GST_IS_BUFFER(buf));

  //usleep(10000);
  jpegenc = GST_JPEGENC(pad->parent);

  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"gst_jpegenc_chain: got buffer of %ld bytes in '%s'\n",size,
          gst_element_get_name(GST_ELEMENT(jpegenc)));

  /* first deal with video metadata */
  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (jpegenc->format != ((MetaVideoRaw *)meta)->format ||
        jpegenc->width  != ((MetaVideoRaw *)meta)->width ||
        jpegenc->height != ((MetaVideoRaw *)meta)->height) {

      jpegenc->format = ((MetaVideoRaw *)meta)->format;
      jpegenc->width  = ((MetaVideoRaw *)meta)->width;
      jpegenc->height = ((MetaVideoRaw *)meta)->height;

      if (jpegenc->buffer) gst_buffer_unref(jpegenc->buffer);
    
      gst_jpegenc_resync(jpegenc);
    }
  }
	
  outbuf = gst_buffer_new();
  outsize = GST_BUFFER_SIZE(outbuf) = jpegenc->bufsize;
  outdata = GST_BUFFER_DATA(outbuf) = g_malloc(outsize);
  GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf);

  width = jpegenc->width;
  height = jpegenc->height;

  base[0] = data;
  base[1] = base[0]+width*height;
  base[2] = base[1]+width*height/4;

  jpegenc->jdest.next_output_byte = outdata;
  jpegenc->jdest.free_in_buffer = outsize;

  jpeg_start_compress(&jpegenc->cinfo, TRUE);

  width2 = width>>1;
  GST_DEBUG (0,"gst_jpegdec_chain: compressing\n");

  for (i = 0; i < height; i += 2*DCTSIZE) {
    for (j=0, k=0; j<2*DCTSIZE;j+=2, k++) {
      jpegenc->line[0][j]   = base[0]; base[0] += width;
      jpegenc->line[0][j+1] = base[0]; base[0] += width;
      jpegenc->line[1][k]   = base[1]; base[1] += width2;
      jpegenc->line[2][k]   = base[2]; base[2] += width2;
    }
    jpeg_write_raw_data(&jpegenc->cinfo, jpegenc->line, 2*DCTSIZE);
  }
  jpeg_finish_compress(&jpegenc->cinfo);
  GST_DEBUG (0,"gst_jpegdec_chain: compressing done\n");

  GST_BUFFER_SIZE(outbuf) = (((outsize - jpegenc->jdest.free_in_buffer)+3)&~3);

  gst_pad_push(jpegenc->srcpad, outbuf);

  gst_buffer_unref(buf);
}
