/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#include <sys/soundcard.h>

#include "gstlame.h"


GstElementDetails gst_lame_details = {
  "L.A.M.E. mp3 encoder",
  "Filter/Encoder/Audio",
  "High-quality free MP3 encoder",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 2000",
};

static GstPadFactory gst_lame_sink_factory = {
  "sink",
  GST_PAD_FACTORY_SINK,
  GST_PAD_FACTORY_ALWAYS,
  GST_PAD_FACTORY_CAPS (
  "gstlame_sink",
    "audio/raw",
    "format",     GST_PROPS_INT (AFMT_S16_LE),
    "depth",      GST_PROPS_INT (16),
    "rate",       GST_PROPS_INT_RANGE (8000, 48000),
    "channels",   GST_PROPS_INT_RANGE (1, 2),
    "interleave", GST_PROPS_LIST (
                      GST_PROPS_BOOLEAN (TRUE)
                  )
  ),
  NULL
};

static GstPadFactory gst_lame_src_factory = {
  "src",
  GST_PAD_FACTORY_SRC,
  GST_PAD_FACTORY_ALWAYS,
  GST_PAD_FACTORY_CAPS (
  "gstlame_src",
    "audio/mp3",
    "layer",   GST_PROPS_INT (3),
    "bitrate", GST_PROPS_INT_RANGE (8, 320)
  ),
  NULL,
};

/********** Define useful types for non-programmatic interfaces **********/
#define GST_TYPE_LAME_MODE (gst_lame_mode_get_type())
static GtkType
gst_lame_mode_get_type(void) {
  static GtkType lame_mode_type = 0;
  static GtkEnumValue lame_modes[] = {
    { 0, "0", "Stereo" },
    { 1, "1", "Joint-Stereo" },
    { 2, "2", "Dual channel" },
    { 3, "3", "Mono" },
    { 0, NULL, NULL },
  };
  if (!lame_mode_type) {
    lame_mode_type = gtk_type_register_enum ("GstLameMode", lame_modes);
  }
  return lame_mode_type;
}

#define GST_TYPE_LAME_QUALITY (gst_lame_quality_get_type())
static GtkType
gst_lame_quality_get_type(void) {
  static GtkType lame_quality_type = 0;
  static GtkEnumValue lame_quality[] = {
    { 0, "0", "0 - Best" },
    { 1, "1", "1" },
    { 2, "2", "2" },
    { 3, "3", "3" },
    { 4, "4", "4" },
    { 5, "5", "5 - Default" },
    { 6, "6", "6" },
    { 7, "7", "7" },
    { 8, "8", "8" },
    { 9, "9", "9 - Worst" },
    { 0, NULL, NULL },
  };
  if (!lame_quality_type) {
    lame_quality_type = gtk_type_register_enum ("GstLameQuality", lame_quality);
  }
  return lame_quality_type;
}

#define GST_TYPE_LAME_PADDING (gst_lame_padding_get_type())
static GtkType
gst_lame_padding_get_type(void) {
  static GtkType lame_padding_type = 0;
  static GtkEnumValue lame_padding[] = {
    { 0, "0", "No Padding" },
    { 1, "1", "Always Pad" },
    { 2, "2", "Adjust Padding" },
    { 0, NULL, NULL },
  };
  if (!lame_padding_type) {
    lame_padding_type = gtk_type_register_enum ("GstLamePadding", lame_padding);
  }
  return lame_padding_type;
}

/********** Standard stuff for signals and arguments **********/
/* GstLame signals and args */
enum {
  /* FILL_ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_BITRATE,
  ARG_COMPRESSION_RATIO,
  ARG_QUALITY,
  ARG_MODE,
  ARG_MODE_FIXED,
  ARG_FORCE_MS,
  ARG_FREE_FORMAT,
  ARG_COPYRIGHT,
  ARG_ORIGINAL,
  ARG_ERROR_PROTECTION,
  ARG_PADDING_TYPE,
  ARG_EXTENSION,
  ARG_STRICT_ISO,
  ARG_DISABLE_RESERVOIR,
  ARG_VBR,
  ARG_VBR_MEAN_BITRATE,
  ARG_VBR_MIN_BITRATE,
  ARG_VBR_MAX_BITRATE,
  ARG_VBR_HARD_MIN,
  ARG_LOWPASS_FREQ,
  ARG_LOWPASS_WIDTH,
  ARG_HIGHPASS_FREQ,
  ARG_HIGHPASS_WIDTH,
  ARG_ATH_ONLY,
  ARG_ATH_SHORT,
  ARG_NO_ATH,
//  ARG_ATH_TYPE,		// note: CVS has this, 3.87 doesn't
  ARG_ATH_LOWER,
  ARG_CWLIMIT,
  ARG_ALLOW_DIFF_SHORT,
  ARG_NO_SHORT_BLOCKS,
  ARG_EMPHASIS,
};


static void 			gst_lame_class_init	(GstLameClass *klass);
static void 			gst_lame_init		(GstLame *gst_lame);

static void 			gst_lame_set_arg	(GtkObject *object, GtkArg *arg, guint id);
static void 			gst_lame_get_arg	(GtkObject *object, GtkArg *arg, guint id);

static void 			gst_lame_chain		(GstPad *pad, GstBuffer *buf);

static GstElementStateReturn 	gst_lame_change_state 	(GstElement *element);


static GstPadTemplate *srctemp, *sinktemp;

static GstElementClass *parent_class = NULL;
//static guint gst_lame_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_lame_get_type (void) 
{
  static GtkType gst_lame_type = 0;

  if (!gst_lame_type) {
    static const GtkTypeInfo gst_lame_info = {
      "GstLame",
      sizeof(GstLame),
      sizeof(GstLameClass),
      (GtkClassInitFunc)gst_lame_class_init,
      (GtkObjectInitFunc)gst_lame_init,
      (GtkArgSetFunc)gst_lame_set_arg,
      (GtkArgGetFunc)gst_lame_get_arg,
      (GtkClassInitFunc)NULL,
    };
    gst_lame_type = gtk_type_unique (GST_TYPE_ELEMENT, &gst_lame_info);
  }
  return gst_lame_type;
}

static void
gst_lame_class_init (GstLameClass *klass) 
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class (GST_TYPE_ELEMENT);

  gtk_object_add_arg_type ("GstLame::bitrate", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_BITRATE);
  gtk_object_add_arg_type ("GstLame::compression_ratio", GTK_TYPE_FLOAT,
                             GTK_ARG_READWRITE, ARG_COMPRESSION_RATIO);
  gtk_object_add_arg_type ("GstLame::quality", GST_TYPE_LAME_QUALITY,
                             GTK_ARG_READWRITE, ARG_QUALITY);
  gtk_object_add_arg_type ("GstLame::mode", GST_TYPE_LAME_MODE,
                             GTK_ARG_READWRITE, ARG_MODE);
  gtk_object_add_arg_type ("GstLame::mode_fixed", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_MODE_FIXED);
  gtk_object_add_arg_type ("GstLame::force_ms", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_FORCE_MS);
  gtk_object_add_arg_type ("GstLame::free_format", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_FREE_FORMAT);
  gtk_object_add_arg_type ("GstLame::copyright", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_COPYRIGHT);
  gtk_object_add_arg_type ("GstLame::original", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_ORIGINAL);
  gtk_object_add_arg_type ("GstLame::error_protection", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_ERROR_PROTECTION);
  gtk_object_add_arg_type ("GstLame::padding_type", GST_TYPE_LAME_PADDING,
                             GTK_ARG_READWRITE, ARG_PADDING_TYPE);
  gtk_object_add_arg_type ("GstLame::extension", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_EXTENSION);
  gtk_object_add_arg_type ("GstLame::strict_iso", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_STRICT_ISO);
  gtk_object_add_arg_type ("GstLame::disable_reservoir", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_DISABLE_RESERVOIR);
  gtk_object_add_arg_type ("GstLame::vbr", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_VBR);
  gtk_object_add_arg_type ("GstLame::vbr_mean_bitrate", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_VBR_MEAN_BITRATE);
  gtk_object_add_arg_type ("GstLame::vbr_min_bitrate", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_VBR_MIN_BITRATE);
  gtk_object_add_arg_type ("GstLame::vbr_max_bitrate", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_VBR_MAX_BITRATE);
  gtk_object_add_arg_type ("GstLame::vbr_hard_min", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_VBR_HARD_MIN);
  gtk_object_add_arg_type ("GstLame::lowpass_freq", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_LOWPASS_FREQ);
  gtk_object_add_arg_type ("GstLame::lowpass_width", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_LOWPASS_WIDTH);
  gtk_object_add_arg_type ("GstLame::highpass_freq", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_HIGHPASS_FREQ);
  gtk_object_add_arg_type ("GstLame::highpass_width", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_HIGHPASS_WIDTH);
  gtk_object_add_arg_type ("GstLame::ath_only", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_ATH_ONLY);
  gtk_object_add_arg_type ("GstLame::ath_short", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_ATH_SHORT);
  gtk_object_add_arg_type ("GstLame::no_ath", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_NO_ATH);
//  gtk_object_add_arg_type ("GstLame::ath_type", GTK_TYPE_INT,
//                             GTK_ARG_READWRITE, ARG_ATH_TYPE);
  gtk_object_add_arg_type ("GstLame::ath_lower", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_ATH_LOWER);
  gtk_object_add_arg_type ("GstLame::cwlimit", GTK_TYPE_INT,
                             GTK_ARG_READWRITE, ARG_CWLIMIT);
  gtk_object_add_arg_type ("GstLame::allow_diff_short", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_ALLOW_DIFF_SHORT);
  gtk_object_add_arg_type ("GstLame::no_short_blocks", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_NO_SHORT_BLOCKS);
  gtk_object_add_arg_type ("GstLame::emphasis", GTK_TYPE_BOOL,
                             GTK_ARG_READWRITE, ARG_EMPHASIS);

  gtkobject_class->set_arg = gst_lame_set_arg;
  gtkobject_class->get_arg = gst_lame_get_arg;

  gstelement_class->change_state = gst_lame_change_state;
}

static void 
gst_lame_init (GstLame *lame) 
{
  GST_DEBUG_ENTER("(\"%s\")",gst_element_get_name(GST_ELEMENT(lame)));

  lame->sinkpad = gst_pad_new_from_template (sinktemp, "sink");
  gst_pad_set_chain_function (lame->sinkpad, gst_lame_chain);
  gst_element_add_pad (GST_ELEMENT (lame), lame->sinkpad);

  lame->srcpad = gst_pad_new_from_template (srctemp, "src");
  gst_element_add_pad (GST_ELEMENT (lame), lame->srcpad);

  GST_DEBUG (0,"setting up lame encoder\n");
  lame_init(&lame->lgf);

  lame->samplerate = 44100;
  lame->num_channels = 2;

  lame->bitrate = lame->lgf.brate;
  lame->compression_ratio = lame->lgf.compression_ratio;
  lame->quality = lame->lgf.quality;
  lame->mode = lame->lgf.mode;
  lame->mode_fixed = lame->lgf.mode_fixed;
  lame->force_ms = lame->lgf.force_ms;
  lame->free_format = lame->lgf.free_format;
  lame->copyright = lame->lgf.copyright;
  lame->original = lame->lgf.original;
  lame->error_protection = lame->lgf.error_protection;
  lame->padding_type = lame->lgf.padding_type;
  lame->extension = lame->lgf.extension;
  lame->strict_iso = lame->lgf.strict_ISO;
  lame->disable_reservoir = lame->lgf.disable_reservoir;
  lame->vbr = lame->lgf.VBR_q;
  lame->vbr_mean_bitrate = lame->lgf.VBR_mean_bitrate_kbps;
  lame->vbr_min_bitrate = lame->lgf.VBR_min_bitrate_kbps;
  lame->vbr_max_bitrate = lame->lgf.VBR_max_bitrate_kbps;
  lame->vbr_hard_min = lame->lgf.VBR_hard_min;
  lame->lowpass_freq = lame->lgf.lowpassfreq;
  lame->lowpass_width = lame->lgf.lowpasswidth;
  lame->highpass_freq = lame->lgf.highpassfreq;
  lame->highpass_width = lame->lgf.highpasswidth;
  lame->ath_only = lame->lgf.ATHonly;
  lame->ath_short = lame->lgf.ATHshort;
  lame->no_ath = lame->lgf.noATH;
//  lame->ath_type = lame->lgf.ATHtype;
  lame->ath_lower = lame->lgf.ATHlower;
  lame->cwlimit = lame->lgf.cwlimit;
  lame->allow_diff_short = lame->lgf.allow_diff_short;
  lame->no_short_blocks = lame->lgf.no_short_blocks;
  lame->emphasis = lame->lgf.emphasis;

  GST_DEBUG (0,"done initializing lame element\n");;
}


static void 
gst_lame_set_arg (GtkObject *object, GtkArg *arg, guint id) 
{
  GstLame *lame;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_LAME (object));
  
  lame = GST_LAME (object);

  switch(id) {
    case ARG_BITRATE:
      lame->bitrate = GTK_VALUE_INT (*arg);
      break;
    case ARG_COMPRESSION_RATIO:
      lame->compression_ratio = GTK_VALUE_FLOAT (*arg);
      break;
    case ARG_QUALITY:
      lame->quality = GTK_VALUE_INT (*arg);
      break;
    case ARG_MODE:
      lame->mode = GTK_VALUE_INT (*arg);
      break;
    case ARG_MODE_FIXED:
      lame->mode_fixed = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_FORCE_MS:
      lame->force_ms = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_FREE_FORMAT:
      lame->free_format = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_COPYRIGHT:
      lame->copyright = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_ORIGINAL:
      lame->original = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_ERROR_PROTECTION:
      lame->error_protection = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_PADDING_TYPE:
      lame->padding_type = GTK_VALUE_INT (*arg);
      break;
    case ARG_EXTENSION:
      lame->extension = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_STRICT_ISO:
      lame->strict_iso = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_DISABLE_RESERVOIR:
      lame->disable_reservoir = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_VBR:
      lame->vbr = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_VBR_MEAN_BITRATE:
      lame->vbr_mean_bitrate = GTK_VALUE_INT (*arg);
      break;
    case ARG_VBR_MIN_BITRATE:
      lame->vbr_min_bitrate = GTK_VALUE_INT (*arg);
      break;
    case ARG_VBR_MAX_BITRATE:
      lame->vbr_max_bitrate = GTK_VALUE_INT (*arg);
      break;
    case ARG_VBR_HARD_MIN:
      lame->vbr_hard_min = GTK_VALUE_INT (*arg);
      break;
    case ARG_LOWPASS_FREQ:
      lame->lowpass_freq = GTK_VALUE_INT (*arg);
      break;
    case ARG_LOWPASS_WIDTH:
      lame->lowpass_width = GTK_VALUE_INT (*arg);
      break;
    case ARG_HIGHPASS_FREQ:
      lame->highpass_freq = GTK_VALUE_INT (*arg);
      break;
    case ARG_HIGHPASS_WIDTH:
      lame->highpass_width = GTK_VALUE_INT (*arg);
      break;
    case ARG_ATH_ONLY:
      lame->ath_only = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_ATH_SHORT:
      lame->ath_short = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_NO_ATH:
      lame->no_ath = GTK_VALUE_BOOL (*arg);
      break;
//    case ARG_ATH_TYPE:
//      lame->ath_type = GTK_VALUE_INT (*arg);
//      break;
    case ARG_ATH_LOWER:
      lame->ath_lower = GTK_VALUE_INT (*arg);
      break;
    case ARG_CWLIMIT:
      lame->cwlimit = GTK_VALUE_INT (*arg);
      break;
    case ARG_ALLOW_DIFF_SHORT:
      lame->allow_diff_short = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_NO_SHORT_BLOCKS:
      lame->no_short_blocks = GTK_VALUE_BOOL (*arg);
      break;
    case ARG_EMPHASIS:
      lame->emphasis = GTK_VALUE_BOOL (*arg);
      break;
    default:
      break;
  }

}

static void 
gst_lame_get_arg (GtkObject *object,GtkArg *arg,guint id) 
{
  GstLame *lame;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_LAME (object));
  
  lame = GST_LAME (object);

  switch (id) {
    case ARG_BITRATE:
      GTK_VALUE_INT (*arg) = lame->bitrate;
      break;
    case ARG_COMPRESSION_RATIO:
      GTK_VALUE_FLOAT (*arg) = lame->compression_ratio;
      break;
    case ARG_QUALITY:
      GTK_VALUE_INT (*arg) = lame->quality;
      break;
    case ARG_MODE:
      GTK_VALUE_INT (*arg) = lame->mode;
      break;
    case ARG_MODE_FIXED:
      GTK_VALUE_BOOL (*arg) = lame->mode_fixed;
      break;
    case ARG_FORCE_MS:
      GTK_VALUE_BOOL (*arg) = lame->force_ms;
      break;
    case ARG_FREE_FORMAT:
      GTK_VALUE_BOOL (*arg) = lame->free_format;
      break;
    case ARG_COPYRIGHT:
      GTK_VALUE_BOOL (*arg) = lame->copyright;
      break;
    case ARG_ORIGINAL:
      GTK_VALUE_BOOL (*arg) = lame->original;
      break;
    case ARG_ERROR_PROTECTION:
      GTK_VALUE_BOOL (*arg) = lame->error_protection;
      break;
    case ARG_PADDING_TYPE:
      GTK_VALUE_INT (*arg) = lame->padding_type;
      break;
    case ARG_EXTENSION:
      GTK_VALUE_BOOL (*arg) = lame->extension;
      break;
    case ARG_STRICT_ISO:
      GTK_VALUE_BOOL (*arg) = lame->strict_iso;
      break;
    case ARG_DISABLE_RESERVOIR:
      GTK_VALUE_BOOL (*arg) = lame->disable_reservoir;
      break;
    case ARG_VBR:
      GTK_VALUE_BOOL (*arg) = lame->vbr;
      break;
    case ARG_VBR_MEAN_BITRATE:
      GTK_VALUE_INT (*arg) = lame->vbr_mean_bitrate;
      break;
    case ARG_VBR_MIN_BITRATE:
      GTK_VALUE_INT (*arg) = lame->vbr_min_bitrate;
      break;
    case ARG_VBR_MAX_BITRATE:
      GTK_VALUE_INT (*arg) = lame->vbr_max_bitrate;
      break;
    case ARG_VBR_HARD_MIN:
      GTK_VALUE_INT (*arg) = lame->vbr_hard_min;
      break;
    case ARG_LOWPASS_FREQ:
      GTK_VALUE_INT (*arg) = lame->lowpass_freq;
      break;
    case ARG_LOWPASS_WIDTH:
      GTK_VALUE_INT (*arg) = lame->lowpass_width;
      break;
    case ARG_HIGHPASS_FREQ:
      GTK_VALUE_INT (*arg) = lame->highpass_freq;
      break;
    case ARG_HIGHPASS_WIDTH:
      GTK_VALUE_INT (*arg) = lame->highpass_width;
      break;
    case ARG_ATH_ONLY:
      GTK_VALUE_BOOL (*arg) = lame->ath_only;
      break;
    case ARG_ATH_SHORT:
      GTK_VALUE_BOOL (*arg) = lame->ath_short;
      break;
    case ARG_NO_ATH:
      GTK_VALUE_BOOL (*arg) = lame->no_ath;
      break;
//    case ARG_ATH_TYPE:
//      GTK_VALUE_INT (*arg) = lame->ath_type;
//      break;
    case ARG_ATH_LOWER:
      GTK_VALUE_INT (*arg) = lame->ath_lower;
      break;
    case ARG_CWLIMIT:
      GTK_VALUE_INT (*arg) = lame->cwlimit;
      break;
    case ARG_ALLOW_DIFF_SHORT:
      GTK_VALUE_BOOL (*arg) = lame->allow_diff_short;
      break;
    case ARG_NO_SHORT_BLOCKS:
      GTK_VALUE_BOOL (*arg) = lame->no_short_blocks;
      break;
    case ARG_EMPHASIS:
      GTK_VALUE_BOOL (*arg) = lame->emphasis;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static void 
gst_lame_chain (GstPad *pad, GstBuffer *buf)
{
  GstLame *lame;
  GstBuffer *outbuf;
  gchar *mp3_data;
  gint mp3_buffer_size, mp3_size;

  g_return_if_fail (pad != NULL);
  lame = GST_LAME (gst_pad_get_parent (pad));

  mp3_buffer_size = ((GST_BUFFER_SIZE(buf) / 4) * 1.25) + 7200;
  mp3_data = g_malloc (mp3_buffer_size);

  mp3_size = lame_encode_buffer_interleaved (&lame->lgf, (short int *)(GST_BUFFER_DATA (buf)),
               GST_BUFFER_SIZE (buf) / 4, mp3_data, mp3_buffer_size);
  GST_DEBUG (0,"encoded %d bytes of audio to %d bytes of mp3\n",GST_BUFFER_SIZE (buf), mp3_size);

  if (mp3_size > 0) {
    outbuf = gst_buffer_new ();
    GST_BUFFER_DATA (outbuf) = mp3_data;
    GST_BUFFER_SIZE (outbuf) = mp3_size;

    gst_pad_push(lame->srcpad,outbuf);
  }
  else g_free (mp3_data);

}

/* transition to the READY state by configuring the gst_lame encoder */
static gboolean 
gst_lame_setup (GstLame *lame) 
{
  GST_DEBUG_ENTER("(\"%s\")",gst_element_get_name(GST_ELEMENT(lame)));

  // copy the parameters over
  lame->lgf.in_samplerate = lame->samplerate;
  if (lame->num_channels == 2) lame->lgf.mode = 3;
  else lame->lgf.mode = lame->mode;

  lame->lgf.brate = lame->bitrate;
  lame->lgf.compression_ratio = lame->compression_ratio;
  lame->lgf.quality = lame->quality;
  lame->lgf.mode = lame->mode;
  lame->lgf.mode_fixed = lame->mode_fixed;
  lame->lgf.force_ms = lame->force_ms;
  lame->lgf.free_format = lame->free_format;
  lame->lgf.copyright = lame->copyright;
  lame->lgf.original = lame->original;
  lame->lgf.error_protection = lame->error_protection;
  lame->lgf.padding_type = lame->padding_type;
  lame->lgf.extension = lame->extension;
  lame->lgf.strict_ISO = lame->strict_iso;
  lame->lgf.disable_reservoir = lame->disable_reservoir;
  lame->lgf.VBR_q = lame->vbr;
  lame->lgf.VBR_mean_bitrate_kbps = lame->vbr_mean_bitrate;
  lame->lgf.VBR_min_bitrate_kbps = lame->vbr_min_bitrate;
  lame->lgf.VBR_max_bitrate_kbps = lame->vbr_max_bitrate;
  lame->lgf.VBR_hard_min = lame->vbr_hard_min;
  lame->lgf.lowpassfreq = lame->lowpass_freq;
  lame->lgf.lowpasswidth = lame->lowpass_width;
  lame->lgf.highpassfreq = lame->highpass_freq;
  lame->lgf.highpasswidth = lame->highpass_width;
  lame->lgf.ATHonly = lame->ath_only;
  lame->lgf.ATHshort = lame->ath_short;
  lame->lgf.noATH = lame->no_ath;
//  lame->lgf.ATHtype = lame->ath_type;
  lame->lgf.ATHlower = lame->ath_lower;
  lame->lgf.cwlimit = lame->cwlimit;
  lame->lgf.allow_diff_short = lame->allow_diff_short;
  lame->lgf.no_short_blocks = lame->no_short_blocks;
  lame->lgf.emphasis = lame->emphasis;

  // initialize the lame encoder
  if (lame_init_params(&lame->lgf) < 0) {
    GST_DEBUG (0,"error initializinglame library\n");
    return FALSE;
  }

  GST_DEBUG_LEAVE("");

  return TRUE;
}

static GstElementStateReturn
gst_lame_change_state (GstElement *element)
{
  g_return_val_if_fail (GST_IS_LAME (element), GST_STATE_FAILURE);

  GST_DEBUG (0,"state pending %d\n", GST_STATE_PENDING (element));

  /* if going down into NULL state, close the file if it's open */
  if (GST_STATE_PENDING (element) == GST_STATE_READY) {
    gst_lame_setup(GST_LAME(element));
  }

  /* if we haven't failed already, give the parent class a chance to ;-) */
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

GstPlugin *
plugin_init (GModule *module) 
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new ("lame");
  g_return_val_if_fail (plugin != NULL, NULL);

  /* create an elementfactory for the gst_lame element */
  factory = gst_elementfactory_new ("lame", GST_TYPE_LAME,
                                    &gst_lame_details);
  g_return_val_if_fail (factory != NULL, NULL);

  /* register the source's padtemplate */
  srctemp = gst_padtemplate_new (&gst_lame_src_factory);
  gst_elementfactory_add_padtemplate (factory, srctemp);

  /* register the sink's padtemplate */
  sinktemp = gst_padtemplate_new (&gst_lame_sink_factory);
  gst_elementfactory_add_padtemplate (factory, sinktemp);

  /* and add the gst_lame element factory to the plugin */
  gst_plugin_add_factory (plugin, factory);
  
  return plugin;
}
