/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <string.h>

//#define DEBUG_ENABLED

#include <inttypes.h>

#include "output_video.h"
#include "gstmpeg2dec.h"

/* elementfactory information */
static GstElementDetails gst_mpeg2dec_details = {
  "mpeg1 and mpeg2 video decoder",
  "Filter/Decoder/Video",
  "Uses modified mpeg2dec to decode MPEG video streams",
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

// This must match decoder and encoder tables
static double video_rate[16] =
{
  0.0,
  24000.0/1001.,
  24.0,
  25.0,
  30000.0/1001.,
  30.0,
  50.0,
  60000.0/1001.,
  60.0,
  1,
  5,
  10,
  12,
  15,
  0,
  0
};


/* Mpeg2dec signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};

static void gst_mpeg2dec_class_init(GstMpeg2decClass *klass);
static void gst_mpeg2dec_init(GstMpeg2dec *mpeg2dec);

static void gst_mpeg2dec_chain (GstPad *pad, GstBuffer *buf);

static void gst_mpeg2dec_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_mpeg2dec_get_arg(GtkObject *object,GtkArg *arg,guint id);

static void gst_mpeg2dec_new_frame (char *buf, void *data, guint64 timestamp);

static GstElementClass *parent_class = NULL;
//static guint gst_mpeg2dec_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_mpeg2dec_get_type (void) 
{
  static GtkType mpeg2dec_type = 0;

  if (!mpeg2dec_type) {
    static const GtkTypeInfo mpeg2dec_info = {
      "GstMpeg2dec",
      sizeof(GstMpeg2dec),
      sizeof(GstMpeg2decClass),
      (GtkClassInitFunc)gst_mpeg2dec_class_init,
      (GtkObjectInitFunc)gst_mpeg2dec_init,
      (GtkArgSetFunc)gst_mpeg2dec_set_arg,
      (GtkArgGetFunc)gst_mpeg2dec_get_arg,
      (GtkClassInitFunc)NULL,
    };
    mpeg2dec_type = gtk_type_unique(GST_TYPE_ELEMENT,&mpeg2dec_info);
  }
  return mpeg2dec_type;
}

static void
gst_mpeg2dec_class_init(GstMpeg2decClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_mpeg2dec_set_arg;
  gtkobject_class->get_arg = gst_mpeg2dec_get_arg;

}

static void 
gst_mpeg2dec_init (GstMpeg2dec *mpeg2dec) 
{
  /* create the sink and src pads */
  mpeg2dec->sinkpad = gst_pad_new ("sink", GST_PAD_SINK);
  gst_element_add_pad (GST_ELEMENT (mpeg2dec), mpeg2dec->sinkpad);
  gst_pad_set_chain_function (mpeg2dec->sinkpad, gst_mpeg2dec_chain);

  mpeg2dec->srcpad = gst_pad_new ("src", GST_PAD_SRC);
  gst_element_add_pad (GST_ELEMENT (mpeg2dec), mpeg2dec->srcpad);

  /* initialize the mpeg2dec decoder state */
  mpeg2dec->decoder = g_new (mpeg2dec_t, 1);
  mpeg2_init (mpeg2dec->decoder, gst_mpeg2dec_new_frame, (void *)mpeg2dec);
  mpeg2dec->next_time = 0;

  // reset the initial video state
  mpeg2dec->format = -1;
  mpeg2dec->width = -1;
  mpeg2dec->height = -1;
  mpeg2dec->meta = NULL;
}

static void 
gst_mpeg2dec_new_frame (char *buf, void *data, guint64 timestamp) 
{
  GstBuffer *outbuf;
  GstMpeg2dec *mpeg2dec;

  mpeg2dec = GST_MPEG2DEC (data);

  outbuf = gst_buffer_new ();
  //GST_BUFFER_SIZE (outbuf) = MPEG2DEC_FRAME_SIZE (mpeg2dec->decoder);
  GST_BUFFER_SIZE (outbuf) = mpeg2dec->decoder->picture->coded_picture_width *
  			     mpeg2dec->decoder->picture->coded_picture_height * 3/2;
  GST_BUFFER_DATA (outbuf) = g_malloc (GST_BUFFER_SIZE (outbuf));
  memcpy (GST_BUFFER_DATA (outbuf), buf, GST_BUFFER_SIZE (outbuf));

  //g_print ("mpeg2dec: timestamp out %llu\n", timestamp);
  
  if (mpeg2dec->meta == NULL) {
    mpeg2dec->meta = gst_meta_new_size (sizeof (MetaVideoRaw));

    ((MetaVideoRaw *)mpeg2dec->meta)->format = GST_COLORSPACE_YUV420P;
    //((MetaVideoRaw *)mpeg2dec->meta)->format = GST_COLORSPACE_BGR565;
    ((MetaVideoRaw *)mpeg2dec->meta)->width = mpeg2dec->decoder->picture->coded_picture_width;
    ((MetaVideoRaw *)mpeg2dec->meta)->height = mpeg2dec->decoder->picture->coded_picture_height;
 
    mpeg2dec->time_interval = 1000000.0/video_rate[mpeg2dec->decoder->picture->frame_rate_code];

  }

  GST_BUFFER_TIMESTAMP (outbuf) = mpeg2dec->next_time;
  //GST_BUFFER_TIMESTAMP (outbuf) = timestamp;

  mpeg2dec->next_time += mpeg2dec->time_interval;

  gst_meta_ref (mpeg2dec->meta);
  gst_buffer_add_meta (outbuf, mpeg2dec->meta);

  gst_pad_push(mpeg2dec->srcpad, outbuf);
}

static void 
gst_mpeg2dec_chain (GstPad *pad, GstBuffer *buf) 
{
  GstMpeg2dec *mpeg2dec;
  gint num_frames;

  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);

  mpeg2dec = GST_MPEG2DEC (pad->parent);

  if (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLUSH))
    //mpeg2dec->decoder->is_picture_needed = 1;
    mpeg2dec->decoder->is_sequence_needed = 1;

  //g_print ("mpeg2dec: timestamp in  %llu\n", GST_BUFFER_TIMESTAMP (buf));

  num_frames = mpeg2_decode_data(mpeg2dec->decoder, 
		                   GST_BUFFER_DATA (buf),
		                   GST_BUFFER_DATA (buf) + GST_BUFFER_SIZE (buf));
    
  gst_buffer_unref(buf);

}

static void 
gst_mpeg2dec_set_arg (GtkObject *object,
		      GtkArg *arg,
		      guint id) 
{
  GstMpeg2dec *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_MPEG2DEC (object));
  src = GST_MPEG2DEC (object);

  switch(id) {
    default:
      break;
  }
}

static void 
gst_mpeg2dec_get_arg (GtkObject *object,
		      GtkArg *arg,
		      guint id) 
{
  GstMpeg2dec *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_MPEG2DEC (object));
  src = GST_MPEG2DEC (object);

  switch(id) {
    default:
      break;
  }
}


GstPlugin*
plugin_init (GModule *module) 
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("mpeg2dec");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the mpeg2dec element */
  factory = gst_elementfactory_new("mpeg2dec",GST_TYPE_MPEG2DEC,
                                   &gst_mpeg2dec_details);
  g_return_val_if_fail(factory != NULL, NULL); 
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
