/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
//#define DEBUG_ENABLED

#include "gstmpeg2enc.h"

/* elementfactory information */
GstElementDetails gst_mpeg2enc_details = {
  "mpeg1 and mpeg2 video encoder",
  "Filter/Encoder/Video",
  "Uses modified mpeg2encode V1.2a to encode MPEG video streams",
  VERSION,
  "(C) 1996, MPEG Software Simulation Group\n" 
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

/* Mpeg2enc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_FPS,
  /* FILL ME */
};

static double video_rates[16] =
{
  0.0,
  24000.0/1001.,
  24.0,
  25.0,
  30000.0/1001.,
  30.0,
  50.0,
  60000.0/1001.,
  60.0,
  1,
  5,
  10,
  12,
  15,
  0,
  0
};


static void gst_mpeg2enc_class_init(GstMpeg2encClass *klass);
static void gst_mpeg2enc_init(GstMpeg2enc *mpeg2enc);

static void gst_mpeg2enc_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_mpeg2enc_get_arg(GtkObject *object,GtkArg *arg,guint id);

static GstElementClass *parent_class = NULL;
//static guint gst_mpeg2enc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_mpeg2enc_get_type(void) {
  static GtkType mpeg2enc_type = 0;

  if (!mpeg2enc_type) {
    static const GtkTypeInfo mpeg2enc_info = {
      "GstMpeg2enc",
      sizeof(GstMpeg2enc),
      sizeof(GstMpeg2encClass),
      (GtkClassInitFunc)gst_mpeg2enc_class_init,
      (GtkObjectInitFunc)gst_mpeg2enc_init,
      (GtkArgSetFunc)gst_mpeg2enc_set_arg,
      (GtkArgGetFunc)gst_mpeg2enc_get_arg,
      (GtkClassInitFunc)NULL,
    };
    mpeg2enc_type = gtk_type_unique(GST_TYPE_ELEMENT,&mpeg2enc_info);
  }
  return mpeg2enc_type;
}

static void
gst_mpeg2enc_class_init(GstMpeg2encClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  gtk_object_add_arg_type("GstMpeg2enc::frames_per_second", GTK_TYPE_DOUBLE,
	                       GTK_ARG_READWRITE, ARG_FPS);

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_mpeg2enc_set_arg;
  gtkobject_class->get_arg = gst_mpeg2enc_get_arg;

}

static void gst_mpeg2enc_init(GstMpeg2enc *mpeg2enc) {
  /* create the sink and src pads */
  mpeg2enc->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(mpeg2enc),mpeg2enc->sinkpad);
  gst_pad_set_chain_function(mpeg2enc->sinkpad,gst_mpeg2enc_chain);
  mpeg2enc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(mpeg2enc),mpeg2enc->srcpad);

  /* initialize the mpeg2enc encoder state */
  mpeg2enc->encoder = mpeg2enc_new_encoder();
  mpeg2enc->next_time = 0;

  // reset the initial video state
  mpeg2enc->format = -1;
  mpeg2enc->width = -1;
  mpeg2enc->height = -1;
}

GstElement *gst_mpeg2enc_new(gchar *name) {
  GstElement *mpeg2enc = GST_ELEMENT(gtk_type_new(GST_TYPE_MPEG2ENC));
  gst_element_set_name(GST_ELEMENT(mpeg2enc),name);
  return mpeg2enc;
}

void gst_mpeg2enc_chain(GstPad *pad,GstBuffer *buf) {
  GstMpeg2enc *mpeg2enc;
  guchar *data;
  gulong size;
  GstBuffer *outbuf;
  GstMeta *meta;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  //g_return_if_fail(GST_IS_BUFFER(buf));

  mpeg2enc = GST_MPEG2ENC(pad->parent);

  /* first deal with video metadata */
  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (mpeg2enc->format != ((MetaVideoRaw *)meta)->format ||
        mpeg2enc->width  != ((MetaVideoRaw *)meta)->width ||
        mpeg2enc->height != ((MetaVideoRaw *)meta)->height) {

      mpeg2enc->format = ((MetaVideoRaw *)meta)->format;
      mpeg2enc->width  = ((MetaVideoRaw *)meta)->width;
      mpeg2enc->height = ((MetaVideoRaw *)meta)->height;

      mpeg2enc->encoder->seq.horizontal_size = mpeg2enc->width;
      mpeg2enc->encoder->seq.display_horizontal_size = mpeg2enc->width;
      mpeg2enc->encoder->seq.vertical_size = mpeg2enc->height;
      mpeg2enc->encoder->seq.display_vertical_size = mpeg2enc->height;
    }
  }

  data = (guchar *)GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);
  GST_DEBUG (0,"gst_mpeg2enc_chain: got buffer of %ld bytes in '%s'\n",size,
          gst_element_get_name(GST_ELEMENT(mpeg2enc)));

  mpeg2enc->state = mpeg2enc_new_picture(mpeg2enc->encoder, data, size, mpeg2enc->state);

  if (mpeg2enc->state & NEW_DATA) {
    outbuf = gst_buffer_new();
    GST_BUFFER_SIZE(outbuf) = mpeg2enc->encoder->pb.newlen;
    GST_BUFFER_DATA(outbuf) = mpeg2enc->encoder->pb.outbase;

    GST_DEBUG (0,"gst_mpeg2enc_chain: pushing buffer %d\n", GST_BUFFER_SIZE(outbuf));
    gst_pad_push(mpeg2enc->srcpad,outbuf);
  }

  gst_buffer_unref(buf);
}

static void gst_mpeg2enc_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstMpeg2enc *src;
  int i;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_MPEG2ENC(object));
  src = GST_MPEG2ENC(object);

  switch(id) {
    case ARG_FPS:
      src->encoder->seq.frame_rate_code = 3; /* default 25 fps */
      for (i=0; i< 16; i++) {
	if (fabs(video_rates[i] - GTK_VALUE_DOUBLE(*arg)) < .001) {
          src->encoder->seq.frame_rate_code = i;
	  gst_info("mpeg2enc: setting framerate for encoding to %g\n", video_rates[i]);
	  break;
	}
      }
      break;
    default:
      break;
  }
}

static void gst_mpeg2enc_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstMpeg2enc *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_MPEG2ENC(object));
  src = GST_MPEG2ENC(object);

  switch(id) {
    case ARG_FPS:
      GTK_VALUE_DOUBLE(*arg) = src->encoder->seq.frame_rate_code;
      break;
    default:
      break;
  }
}


GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* this filter needs the getbits package */
  if (!gst_library_load("gstputbits")) {
    gst_info("mpeg2enc:: could not load support library: 'gstputbits'\n");
    return NULL;
  }
  if (!gst_library_load("gstidct")) {
    gst_info("mpeg2enc:: could not load support library: 'gstidct'\n");
    return NULL;
  }

  plugin = gst_plugin_new("mpeg2enc");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the mpeg2enc element */
  factory = gst_elementfactory_new("mpeg2enc",GST_TYPE_MPEG2ENC,
                                   &gst_mpeg2enc_details);
  g_return_val_if_fail(factory != NULL, NULL); 
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
