/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/time.h>

#define GST_DEBUG_FORCE_DISABLE

#include "videosink.h"
#include "dga.h"

#include "libs/colorspace/gstcolorspace.h"


GstElementDetails gst_videosink_details = {
  "Video sink",
  "Sink/Video",
  "A general video sink",
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

/* videosink signals and args */
enum {
  SIGNAL_FRAME_DISPLAYED,
  LAST_SIGNAL
};


enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  ARG_WIDGET,
  ARG_SCALE,
  ARG_FRAMES_DISPLAYED,
  ARG_FRAME_TIME,
  ARG_XV_ENABLED,
};

static GstPadFactory sink_factory = {
  "sink",
  GST_PAD_FACTORY_SINK,
  GST_PAD_FACTORY_ALWAYS,
  GST_PAD_FACTORY_CAPS(
  "videosink_sink",
    "video/raw"
  ),
  NULL,
};

static void 	gst_videosink_class_init	(GstVideoSinkClass *klass);
static void 	gst_videosink_init		(GstVideoSink *videosink);

//static gboolean gst_videosink_change_state	(GstElement *videosink, GstElementState state);

static void 	gst_videosink_resync		(GstVideoSink *videosink);
static void 	gst_videosink_calc_colorspace	(GstVideoSink *videosink);

static void 	gst_videosink_set_arg		(GtkObject *object,GtkArg *arg,guint id);
static void 	gst_videosink_get_arg		(GtkObject *object,GtkArg *arg,guint id);

static GstPadTemplate *sink_template;

static GstElementClass *parent_class = NULL;
static guint gst_videosink_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_videosink_get_type(void) {
  static GtkType videosink_type = 0;

  if (!videosink_type) {
    static const GtkTypeInfo videosink_info = {
      "GstVideoSink",
      sizeof(GstVideoSink),
      sizeof(GstVideoSinkClass),
      (GtkClassInitFunc)gst_videosink_class_init,
      (GtkObjectInitFunc)gst_videosink_init,
      (GtkArgSetFunc)gst_videosink_set_arg,
      (GtkArgGetFunc)gst_videosink_get_arg,
      (GtkClassInitFunc)NULL,
    };
    videosink_type = gtk_type_unique(GST_TYPE_ELEMENT,&videosink_info);
  }
  return videosink_type;
}

static void 
gst_videosink_class_init (GstVideoSinkClass *klass) 
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class (GST_TYPE_ELEMENT);

  gtk_object_add_arg_type ("GstVideoSink::width", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_WIDTH);
  gtk_object_add_arg_type ("GstVideoSink::height", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_HEIGHT);
  gtk_object_add_arg_type ("GstVideoSink::widget", GTK_TYPE_WIDGET,
                           GTK_ARG_READABLE, ARG_WIDGET);
  gtk_object_add_arg_type ("GstVideoSink::frames_displayed", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_FRAMES_DISPLAYED);
  gtk_object_add_arg_type ("GstVideoSink::frame_time", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_FRAME_TIME);
  gtk_object_add_arg_type ("GstVideoSink::xv_enabled", GTK_TYPE_BOOL,
                           GTK_ARG_READWRITE, ARG_XV_ENABLED);

  gtkobject_class->set_arg = gst_videosink_set_arg;
  gtkobject_class->get_arg = gst_videosink_get_arg;

  gst_videosink_signals[SIGNAL_FRAME_DISPLAYED] =
    gtk_signal_new ("frame_displayed", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (GstVideoSinkClass, frame_displayed),
                    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (gtkobject_class, gst_videosink_signals,
                                  LAST_SIGNAL);

}

void 
gst_videosink_map (GtkWidget *wid, GstVideoSink *sink) 
{
  //g_print("videosink: map\n");
  gst_videosink_resync (sink);
}

void 
gst_videosink_resize (GtkWidget *wid, GtkAllocation *alloc, GstVideoSink *sink) 
{
  g_return_if_fail (sink != NULL);
  g_return_if_fail (GST_IS_VIDEOSINK (sink));

  //g_print("videosink: resize\n");
  if ((sink->width != alloc->width ||
      sink->height != alloc->height)) {
    GST_DEBUG (0,"videosink: doing resize %d %d %d %d\n", alloc->width, alloc->height, alloc->x, alloc->y);
    sink->width = alloc->width;
    sink->height = alloc->height;
    gst_videosink_resync (sink);
  }
}


static void 
gst_videosink_init (GstVideoSink *videosink) 
{
  videosink->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad (GST_ELEMENT (videosink), videosink->sinkpad);
  gst_pad_set_chain_function (videosink->sinkpad, gst_videosink_chain);

  videosink->format = GST_COLORSPACE_RGB32;
  // set the start size...
  videosink->width = 100;
  videosink->height = 100;
  videosink->buffer_width = 100;
  videosink->buffer_height = 100;
  videosink->draw = -1;
  videosink->converter = NULL;
  videosink->scaler = NULL;
  videosink->tempbuffer = NULL;
  videosink->xv_enabled = gdk_xvimage_check_xvideo ();
  videosink->initialized = FALSE;

  gdk_rgb_init ();
  gtk_widget_set_default_colormap (gdk_rgb_get_cmap ());
  //videosink->visual = gdk_rgb_get_visual();
  videosink->visual = gdk_visual_get_system ();
  gst_info("videosink: found %d bits display\n", videosink->visual->depth);
  gtk_widget_set_default_visual (videosink->visual);


  videosink->clock = gst_clock_get_system();
  gst_clock_register(videosink->clock, GST_OBJECT(videosink));

  ((MetaVideoRaw *)&videosink->meta)->overlay_info = NULL;

  GST_FLAG_SET(videosink, GST_ELEMENT_THREAD_SUGGESTED);
}

GstBuffer*
gst_videosink_pull (GstPad *pad) 
{
  GstVideoSink *videosink;
  GstMeta *meta;
  GstBuffer *buf;

  g_return_val_if_fail (pad != NULL, NULL);
  g_return_val_if_fail (GST_IS_PAD (pad), NULL);

  videosink = GST_VIDEOSINK (pad->parent);
  GST_DEBUG (0,"videosink: request for buffer pull %p\n", GST_BUFFER_DATA (videosink->buffer));
  //fprintf(stderr, "videosink: request for buffer pull\n");

  buf = videosink->buffer = gst_buffer_new ();
  GST_BUFFER_DATA (buf) = GTK_OVERLAY_IMAGE_MEM (videosink->image),
    //fprintf(stderr, "videosink: new buffer %p\n", GST_BUFFER_DATA(buf));

  meta = (GstMeta *)&videosink->meta;

  GST_BUFFER_SIZE (buf) = videosink->width * videosink->height * videosink->colorbytes;

  ((MetaVideoRaw *)meta)->format = videosink->colorspace;
  ((MetaVideoRaw *)meta)->width = videosink->width;
  ((MetaVideoRaw *)meta)->height = videosink->height;
  ((MetaVideoRaw *)meta)->overlay_info = &GTK_OVERLAY_IMAGE(videosink->image)->meta_overlay;

  gst_buffer_add_meta (buf, meta);

  gst_buffer_ref (videosink->buffer);

  return videosink->buffer;
}

void 
gst_videosink_chain (GstPad *pad, GstBuffer *buf) 
{
  GstVideoSink *videosink;
  MetaVideoRaw *meta;
  gint draw_format = -1;
  GstClockTimeDiff jitter;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);


  videosink = GST_VIDEOSINK(pad->parent);

  gdk_threads_enter();
  GST_DEBUG (0,"videosink: have buffer of size %d\n", GST_BUFFER_SIZE(buf));

  /* first deal with video metadata */
  meta = (MetaVideoRaw *) gst_buffer_get_first_meta(buf);
  if (meta) {
    if ((meta->format != videosink->buffer_format) ||
  	    (meta->width != videosink->buffer_width) ||
  	    (meta->height != videosink->buffer_height)) {

      videosink->buffer_format = meta->format;
      videosink->buffer_width = meta->width;
      videosink->buffer_height = meta->height;

      videosink->format = meta->format;
      videosink->width = meta->width;
      videosink->height = meta->height;

      gst_videosink_resync(videosink);

    }
    // if we received an overlay buffer
    //  fprintf(stderr, "videosink: received overlay\n");
    if (meta->overlay_info && meta->overlay_info->did_overlay) {
      GST_DEBUG (0,"videosink: received overlay\n");
      gtk_object_set(GTK_OBJECT(videosink->image),"garbage_collect",TRUE, NULL);
      gtk_object_set(GTK_OBJECT(videosink->image),"drawing_enabled",FALSE, NULL);

      if (GTK_OVERLAY_IMAGE_MEM(videosink->image) != GST_BUFFER_DATA(buf)) {
    	goto end;
      }
    }
  }


  if (videosink->xv_enabled) {
    memcpy(videosink->gdk_xvimage->mem, GST_BUFFER_DATA(buf), MIN(GST_BUFFER_SIZE(buf),
			    videosink->buffer_width*videosink->buffer_height+
			    videosink->buffer_width*videosink->buffer_height/2));
  }
  else {

    GST_DEBUG (0,"videosink: trying convert %d to %d\n", videosink->format, videosink->colorspace);

    if (videosink->format == videosink->colorspace) {
      //g_assert(GST_BUFFER_SIZE(buf) >= (videosink->width * videosink->height * videosink->colorbytes));
      if (GTK_OVERLAY_IMAGE_MEM(videosink->image) != GST_BUFFER_DATA(buf)) {
        //fprintf(stderr, "videosink: need copy\n");
        memcpy(GTK_OVERLAY_IMAGE_MEM(videosink->image), GST_BUFFER_DATA(buf), MIN(GST_BUFFER_SIZE(buf),
                GTK_WIDGET(videosink->image)->allocation.width*
		GTK_WIDGET(videosink->image)->allocation.height*2));
      }
      else {
        //fprintf(stderr, "videosink: got buffer back\n");
      }
    }
    else {

      if (videosink->converter == NULL) {

        if (GTK_OVERLAY_IMAGE_TYPE(videosink->image) == GDK_IMAGE_SHARED) {
          // find direct conversion
          videosink->converter = gst_colorspace_converter_new(videosink->buffer_width, videosink->buffer_height, videosink->format, 
			videosink->colorspace, videosink->visual);
          videosink->draw = videosink->colorspace;
          videosink->outbuf = GTK_OVERLAY_IMAGE_MEM(videosink->image);
        }
        else {
          // we have no usefull buffer for conversion
          videosink->outbuf = NULL;
        }

        // try RGB32
        if (videosink->converter == NULL) {
          videosink->converter = gst_colorspace_converter_new(videosink->buffer_width, videosink->buffer_height, 
			videosink->format, GST_COLORSPACE_RGB32, videosink->visual);
          videosink->draw = GST_COLORSPACE_RGB32;
        }
        // try RGB24
        if (videosink->converter == NULL) {
          videosink->converter = gst_colorspace_converter_new(videosink->buffer_width, videosink->buffer_height, 
			videosink->format, GST_COLORSPACE_RGB24, videosink->visual);
          videosink->draw = GST_COLORSPACE_RGB24;
        }
      }

      if (videosink->converter) {
        GST_DEBUG (0,"videosink: converting %d to %d %p\n", videosink->format, videosink->draw, GTK_OVERLAY_IMAGE_MEM(videosink->image));

        if (videosink->scaler) {
          gst_colorspace_convert(videosink->converter, GST_BUFFER_DATA(buf), videosink->tempbuffer);
          gst_videoscale_scale(videosink->scaler, videosink->tempbuffer, GTK_OVERLAY_IMAGE_MEM(videosink->image));
        }
        else {
          if (videosink->outbuf == NULL) videosink->outbuf = g_malloc(videosink->converter->outsize);
          gst_colorspace_convert(videosink->converter, GST_BUFFER_DATA(buf), videosink->outbuf);
        }
      }
      else {
        fprintf(stderr, "videosink: cannot convert %d to %d\n", videosink->format, GST_COLORSPACE_RGB24);
      }
      draw_format = videosink->draw;
    }
  }

  if (!GST_BUFFER_FLAG_IS_SET(buf, GST_BUFFER_FLUSH)) {
    GST_DEBUG (0,"videosink: clock wait: %llu\n", GST_BUFFER_TIMESTAMP(buf));
    videosink->frame_time = GST_BUFFER_TIMESTAMP(buf);
    jitter = gst_clock_current_diff(videosink->clock, videosink->frame_time);
  //printf("videosink: jitter %08lld\n", jitter);
    //if (jitter > 500000) 
    if (0)
    {
      printf("videosink: *************** dropping frame  %llu\n", GST_BUFFER_TIMESTAMP(buf));
      //gst_pad_handle_qos(videosink->sinkpad, jitter);
      goto end;
    }
    else {
      gdk_threads_leave();
      gst_clock_wait(videosink->clock, GST_BUFFER_TIMESTAMP(buf), GST_OBJECT(videosink));
      gdk_threads_enter();
      //gst_pad_handle_qos(videosink->sinkpad, jitter);
    }
    GST_DEBUG (0,"videosink: clock done: %llu\n", GST_BUFFER_TIMESTAMP(buf));
  }
	
  //gdk_threads_enter();
  if (draw_format == GST_COLORSPACE_RGB24) {
    GST_DEBUG (0,"videosink: draw 24 bit\n");
    gdk_draw_rgb_image(videosink->image->window,
		   videosink->image->style->fg_gc[videosink->image->state],
		   0,0,
		   videosink->width,
		   videosink->height,
		   GDK_RGB_DITHER_NORMAL,
		   videosink->outbuf,
		   videosink->width*3);
  }
  else if (draw_format == GST_COLORSPACE_RGB32) {
    GST_DEBUG (0,"videosink: draw 32 bit\n");
    gdk_draw_rgb_32_image(videosink->image->window,
		   videosink->image->style->fg_gc[videosink->image->state],
		   0,0,
	   	   videosink->width,
		   videosink->height,
		   GDK_RGB_DITHER_NORMAL,
		   videosink->outbuf,
		   videosink->width*4);
  }
  else {
    GdkRectangle update_rect;

    update_rect.x = 0;
    update_rect.y = 0;
    update_rect.width = videosink->width;
    update_rect.height = videosink->height;

    if (videosink->xv_enabled) {
      gtk_widget_draw(GTK_WIDGET(videosink->xvimage), &update_rect);
    }
    else if (GTK_IS_OVERLAY_IMAGE(videosink->image)) {
      GST_DEBUG (0,"videosink: draw %p \n", videosink->image);
      gtk_widget_draw(GTK_WIDGET(videosink->image), &update_rect);
      GST_DEBUG (0,"videosink: draw done %p \n", videosink->image);
    }
  }
  //gdk_threads_leave();

  videosink->frames_displayed++;
  GST_DEBUG (0,"videosink: emit new frame \n");
  gdk_threads_leave();
  gtk_signal_emit(GTK_OBJECT(videosink),gst_videosink_signals[SIGNAL_FRAME_DISPLAYED]);
  gdk_threads_enter();
  GST_DEBUG (0,"videosink: emit new frame done\n");

end:
  gdk_threads_leave();

  gst_buffer_unref(buf);
}

/*
static gboolean gst_videosink_change_state(GstElement *element,
                           GstElementState state) {
  GstVideoSink *videosink;

  g_return_val_if_fail(GST_IS_VIDEOSINK(element), FALSE);
  videosink = GST_VIDEOSINK(element);

  switch (state) {
    case GST_STATE_RUNNING:
      break;
    case ~GST_STATE_RUNNING:
      break;
    default:
      break;
  }

  if (GST_ELEMENT_CLASS(parent_class)->change_state)
    return GST_ELEMENT_CLASS(parent_class)->change_state(element,state);
  return TRUE;
}
*/

static void 
gst_videosink_calc_colorspace (GstVideoSink *videosink) 
{
  GdkByteOrder order = GTK_OVERLAY_IMAGE_BYTE_ORDER(videosink->image);

  switch (GTK_OVERLAY_IMAGE_PIXEL_SIZE(videosink->image)) {
    case 16:
      switch (GTK_OVERLAY_IMAGE_DEPTH(videosink->image)) {
        case 15:
          videosink->colorspace = (order == GDK_LSB_FIRST ? GST_COLORSPACE_BGR555: GST_COLORSPACE_RGB555);
          videosink->colorbytes = 2;
          break;
        case 16:
          videosink->colorspace = (order == GDK_LSB_FIRST ? GST_COLORSPACE_BGR565: GST_COLORSPACE_RGB565);
          videosink->colorbytes = 2;
          break;
      }
      break;
    case 24:
      videosink->colorspace = (order == GDK_LSB_FIRST ? GST_COLORSPACE_BGR24 : GST_COLORSPACE_RGB24);
      videosink->colorbytes = 3;
      break;
    case 32:
      videosink->colorspace = (order == GDK_LSB_FIRST ? GST_COLORSPACE_BGR32 : GST_COLORSPACE_RGB32);
      videosink->colorbytes = 4;
      break;
    default:
      videosink->colorspace = -1;
      break;
  }
  GST_DEBUG (0,"videosink: colorspace %d\n", videosink->colorspace);
}

static void 
gst_videosink_resync (GstVideoSink *videosink) 
{
  g_return_if_fail (videosink != NULL);
  g_return_if_fail (GST_IS_VIDEOSINK (videosink));

  if (!videosink->initialized) return;
  
  GST_DEBUG (0,"videosink: doing resync %d %d\n", videosink->width, videosink->height);
  if (videosink->xv_enabled) {
    if (videosink->gdk_xvimage)
      gdk_xvimage_destroy(videosink->gdk_xvimage);
    videosink->gdk_xvimage = gdk_xvimage_new(0x30323449, videosink->buffer_width, videosink->buffer_height);
    gtk_xvimage_set(GTK_XVIMAGE(videosink->xvimage), videosink->gdk_xvimage);
  }
  else {
    gtk_overlay_image_resize(GTK_OVERLAY_IMAGE(videosink->image), videosink->width, videosink->height);
	  
    gst_videosink_calc_colorspace(videosink);

    videosink->tempbuffer = g_realloc(videosink->tempbuffer, videosink->buffer_width*videosink->buffer_height*videosink->colorbytes);
    if (videosink->width != videosink->buffer_width ||
        videosink->height != videosink->buffer_height) {
      GST_DEBUG (0,"videosink: new scaler %d %d %d %d\n", videosink->buffer_width, videosink->buffer_height, videosink->width, videosink->height);
      videosink->scaler = gst_videoscale_new(videosink->buffer_width, videosink->buffer_height, 
		      videosink->width, videosink->height, videosink->colorspace, GST_VIDEOSCALE_NEAREST);
      GST_DEBUG (0,"videosink: new scaler done\n");
    }
    else {
      videosink->scaler = NULL;
    }
    videosink->converter = NULL;
  }
}

static void 
gst_videosink_set_arg (GtkObject *object, GtkArg *arg, guint id) 
{
  GstVideoSink *videosink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_VIDEOSINK (object));
  
  videosink = GST_VIDEOSINK (object);

  switch(id) {
    case ARG_WIDTH:
      videosink->width = GTK_VALUE_INT(*arg);
      GST_DEBUG (0,"videosink: setting width to %d\n", videosink->width);
      gst_videosink_resync(videosink);
      break;
    case ARG_HEIGHT:
      videosink->height = GTK_VALUE_INT(*arg);
      GST_DEBUG (0,"videosink: setting height to %d\n", videosink->height);
      gst_videosink_resync(videosink);
      break;
    case ARG_FRAMES_DISPLAYED: 
      videosink->frames_displayed = GTK_VALUE_INT(*arg);
      break;
    case ARG_FRAME_TIME: 
      videosink->frame_time = GTK_VALUE_INT(*arg);
      break;
    case ARG_XV_ENABLED: 
      videosink->xv_enabled = GTK_VALUE_BOOL(*arg);
      break;
    default:
      break;
  }
}

static void 
gst_videosink_get_arg (GtkObject *object, GtkArg *arg, guint id) 
{
  GstVideoSink *videosink;

  /* it's not null if we got it, but it might not be ours */
  videosink = GST_VIDEOSINK(object);

  switch(id) {
    case ARG_WIDTH: {
      GTK_VALUE_INT(*arg) = videosink->width;
      break;
    }
    case ARG_HEIGHT: {
      GTK_VALUE_INT(*arg) = videosink->height;
      break;
    }
    case ARG_WIDGET: {
      // create the widget
      if (videosink->xv_enabled) {
        gdk_xvimage_init();
        videosink->gdk_xvimage = gdk_xvimage_new(0x32315659, videosink->buffer_width, videosink->buffer_height);
        videosink->xvimage = gtk_xvimage_new(videosink->gdk_xvimage);
        //gtk_widget_show(videosink->xvimage);
        GTK_VALUE_OBJECT(*arg) = GTK_OBJECT(videosink->xvimage);
      }
      else {
        videosink->image = gtk_overlay_image_new();
        //gtk_widget_show(videosink->image);
        gst_info("videosink: using %d bits image %d bits/pixel\n", GTK_OVERLAY_IMAGE_DEPTH(videosink->image),
		                       GTK_OVERLAY_IMAGE_PIXEL_SIZE(videosink->image));
        gst_videosink_calc_colorspace(videosink);

        gtk_signal_connect(GTK_OBJECT(videosink->image),"size-allocate",
            GTK_SIGNAL_FUNC(gst_videosink_resize),videosink);
        gtk_signal_connect(GTK_OBJECT(videosink->image),"map",
            GTK_SIGNAL_FUNC(gst_videosink_map),videosink);
        GTK_VALUE_OBJECT(*arg) = GTK_OBJECT(videosink->image);
      }
      videosink->initialized = TRUE;
      break;
    }
    case ARG_FRAMES_DISPLAYED: {
      GTK_VALUE_INT(*arg) = videosink->frames_displayed;
      break;
    }
    case ARG_FRAME_TIME: {
      GTK_VALUE_INT(*arg) = videosink->frame_time/1000000;
      break;
    }
    case ARG_XV_ENABLED: {
      GTK_VALUE_INT(*arg) = videosink->xv_enabled;
      break;
    }
    default: {
      arg->type = GTK_TYPE_INVALID;
      break;
    }
  }
}



GstPlugin*
plugin_init (GModule *module) 
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* this filter needs the colorspace functions */
  if (!gst_library_load("gstcolorspace")) {
    gst_info("videosink: could not load support library: 'gstcolorspace'\n");
    return NULL;
  }
  if (!gst_library_load("gstvideoscale")) {
    gst_info("videosink: could not load support library: 'gstvideoscale'\n");
    return NULL;
  }

  plugin = gst_plugin_new("videosink");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the videosink element */
  factory = gst_elementfactory_new("videosink",GST_TYPE_VIDEOSINK,
                                   &gst_videosink_details);
  g_return_val_if_fail(factory != NULL, NULL);

  sink_template = gst_padtemplate_new (&sink_factory);
  gst_elementfactory_add_padtemplate (factory, sink_template);

  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
