/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_VIDEOSINK_H__
#define __GST_VIDEOSINK_H__


#include <gtk/gtk.h>
#include <gst/gst.h>
#include <gst/gstclock.h>
#include <gst/meta/videoraw.h>
#include <libs/videoscale/gstvideoscale.h>

#include "config.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/Xmd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>
#include <X11/Xaw/Simple.h>

#ifdef  HAVE_LIBXV
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
#endif

#include "gtkxvimage.h"
#include "gtkoverlayimage.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GST_TYPE_VIDEOSINK \
  (gst_videosink_get_type())
#define GST_VIDEOSINK(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_VIDEOSINK,GstVideoSink))
#define GST_VIDEOSINK_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_VIDEOSINK,GstVideoSink))
#define GST_IS_VIDEOSINK(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_VIDEOSINK))
#define GST_IS_VIDEOSINK_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_VIDEOSINK))

#define MAX_FLIP_BUFFERS 1

typedef struct _GstVideoSink GstVideoSink;
typedef struct _GstVideoSinkClass GstVideoSinkClass;

struct _GstVideoSink {
  GstElement element;

  GstPad *sinkpad;

  gint format, width, height;
  gint buffer_format, buffer_width, buffer_height;
  GtkWidget *image;
  GdkXvImage *gdk_xvimage;
  GtkWidget *xvimage;
  gboolean xv_enabled;
  gboolean initialized;

  gint frames_displayed;
  guint64 frame_time;

  GdkVisual *visual;

  GstClock *clock;

  GstColorSpaceType colorspace;
  GstColorSpaceType draw;
  GstColorSpaceConverter *converter;
  GstVideoScale *scaler;
  guchar *outbuf, *tempbuffer;
  
  guint colorbytes;

  // a buffer with the image of the sink
  GstBuffer *buffer;
  // the meta data for this buffer
  MetaVideoRaw meta;
  // the meta data for DGA
  MetaDGA meta_dga;
};

struct _GstVideoSinkClass {
  GstElementClass parent_class;

  /* signals */
  void (*frame_displayed) (GstElement *element, GstPad *pad);
};

GtkType gst_videosink_get_type(void);
GstElement *gst_videosink_new(gchar *name);
void gst_videosink_chain(GstPad *pad,GstBuffer *buf);
GstBuffer *gst_videosink_pull(GstPad *pad);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_SMOOTHWAVE_H__ */
