/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>

#include "smoothwave.h"

GstElementDetails gst_smoothwave_details = {
  "Smooth waveform",
  "Sink/Visualization",
  "Fading grayscale waveform display",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* SmoothWave signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  ARG_WIDGET,
};


static void gst_smoothwave_class_init(GstSmoothWaveClass *klass);
static void gst_smoothwave_init(GstSmoothWave *smoothwave);

static void gst_smoothwave_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_smoothwave_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstElementClass *parent_class = NULL;
//static guint gst_smoothwave_signals[LAST_SIGNAL] = { 0 };


GtkType
gst_smoothwave_get_type(void) {
  static GtkType smoothwave_type = 0;

  if (!smoothwave_type) {
    static const GtkTypeInfo smoothwave_info = {
      "GstSmoothWave",
      sizeof(GstSmoothWave),
      sizeof(GstSmoothWaveClass),
      (GtkClassInitFunc)gst_smoothwave_class_init,
      (GtkObjectInitFunc)gst_smoothwave_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    smoothwave_type = gtk_type_unique(GST_TYPE_ELEMENT,&smoothwave_info);
  }
  return smoothwave_type;
}

static void gst_smoothwave_class_init(GstSmoothWaveClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstSmoothWave::width", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_WIDTH);
  gtk_object_add_arg_type("GstSmoothWave::height", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_HEIGHT);
  gtk_object_add_arg_type("GstSmoothWave::widget", GTK_TYPE_WIDGET,
                          GTK_ARG_READABLE, ARG_WIDGET);

  gtkobject_class->set_arg = gst_smoothwave_set_arg;
  gtkobject_class->get_arg = gst_smoothwave_get_arg;
}

static void gst_smoothwave_init(GstSmoothWave *smoothwave) {
  int i;
  guint32 palette[256];

  smoothwave->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(smoothwave),smoothwave->sinkpad);
  gst_pad_set_chain_function(smoothwave->sinkpad,gst_smoothwave_chain);
  smoothwave->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(smoothwave),smoothwave->srcpad);

//  smoothwave->meta = NULL;
  smoothwave->width = 512;
  smoothwave->height = 256;

  gdk_rgb_init();
//  gtk_widget_set_default_colormap (gdk_rgb_get_cmap());
//  gtk_widget_set_default_visual (gdk_rgb_get_visual());

//  GST_DEBUG (0,"creating palette\n");
  for (i=0;i<256;i++)
    palette[i] = (i << 16) || (i << 8);
//  GST_DEBUG (0,"creating cmap\n");
  smoothwave->cmap = gdk_rgb_cmap_new(palette,256);
//  GST_DEBUG (0,"created cmap\n");
//  gtk_widget_set_default_colormap (smoothwave->cmap);

  smoothwave->image = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(smoothwave->image),
                        smoothwave->width,smoothwave->height);
  gtk_widget_show(smoothwave->image);

  smoothwave->imagebuffer = g_malloc(smoothwave->width*smoothwave->height);
  memset(smoothwave->imagebuffer,0,smoothwave->width*smoothwave->height);
}

GstElement *gst_smoothwave_new(gchar *name) {
  GstElement *smoothwave = GST_ELEMENT(gtk_type_new(GST_TYPE_SMOOTHWAVE));
  gst_element_set_name(GST_ELEMENT(smoothwave),name);
  return smoothwave;
}

void gst_smoothwave_chain(GstPad *pad,GstBuffer *buf) {
  GstSmoothWave *smoothwave;
  gint16 *samples;
  gint samplecount,i;
  register guint32 *ptr;
  gint qheight;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
//  g_return_if_fail(GST_IS_BUFFER(buf));

  smoothwave = GST_SMOOTHWAVE(pad->parent);

  /* first deal with audio metadata */
//  if (buf->meta) {
//    if (smoothwave->meta != NULL) {
//      /* FIXME: need to unref the old metadata so it goes away */
//    }
//    /* we just make a copy of the pointer */
//    smoothwave->meta = (MetaAudioRaw *)(buf->meta);
//    /* FIXME: now we have to ref the metadata so it doesn't go away */
//  }

//  g_return_if_fail(smoothwave->meta != NULL);

  samples = (gint16 *)GST_BUFFER_DATA(buf);
//  samplecount = buf->datasize / (smoothwave->meta->channels * sizeof(gint16));
  samplecount = GST_BUFFER_SIZE(buf) / (2 * sizeof(gint16));

  qheight = smoothwave->height/4;

//  GST_DEBUG (0,"traversing %d\n",smoothwave->width);
  for (i=0;i<MAX(smoothwave->width,samplecount);i++) {
    gint16 y1 = (gint32)(samples[i*2] * qheight) / 32768 + 
                qheight;
    gint16 y2 = (gint32)(samples[(i*2)+1] * qheight) / 32768 + 
                (qheight*3);
    smoothwave->imagebuffer[y1*smoothwave->width + i] = 0xff;
    smoothwave->imagebuffer[y2*smoothwave->width + i] = 0xff;
//    smoothwave->imagebuffer[i+(smoothwave->width*5)] = i;
  }

  ptr = (guint32 *)smoothwave->imagebuffer;
  for (i=0;i<(smoothwave->width*smoothwave->height)/4;i++) {
    if (*ptr)
      *(ptr++) -= ((*ptr & 0xf0f0f0f0ul) >> 4) + ((*ptr & 0xe0e0e0e0ul) >> 5);
    else
      ptr++;
  }

//  GST_DEBUG (0,"drawing\n");
/*  GST_DEBUG (0,"gdk_draw_indexed_image(%p,%p,%d,%d,%d,%d,%s,%p,%d,%p);\n",
        smoothwave->image->window,
	smoothwave->image->style->fg_gc[GTK_STATE_NORMAL],
	0,0,smoothwave->width,smoothwave->height,
	"GDK_RGB_DITHER_NORMAL",
	smoothwave->imagebuffer,smoothwave->width,
	smoothwave->cmap);*/
/*  gdk_draw_indexed_image(smoothwave->image->window,
	smoothwave->image->style->fg_gc[GTK_STATE_NORMAL],
	0,0,smoothwave->width,smoothwave->height,
	GDK_RGB_DITHER_NONE,
	smoothwave->imagebuffer,smoothwave->width,
	smoothwave->cmap);*/
  gdk_draw_gray_image(smoothwave->image->window,
	smoothwave->image->style->fg_gc[GTK_STATE_NORMAL],
	0,0,smoothwave->width,smoothwave->height,
	GDK_RGB_DITHER_NORMAL,
	smoothwave->imagebuffer,smoothwave->width);

//  gst_trace_add_entry(NULL,0,buf,"smoothwave: calculated smoothwave");

  gst_buffer_unref(buf);
}

static void gst_smoothwave_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstSmoothWave *smoothwave;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_SMOOTHWAVE(object));
  smoothwave = GST_SMOOTHWAVE(object);

  switch(id) {
    case ARG_WIDTH:
      smoothwave->width = GTK_VALUE_INT(*arg);
      gtk_drawing_area_size(GTK_DRAWING_AREA(smoothwave->image),
                            smoothwave->width,smoothwave->height);
      gtk_widget_set_usize(GTK_WIDGET(smoothwave->image),
                           smoothwave->width,smoothwave->height);
      break;
    case ARG_HEIGHT:
      smoothwave->height = GTK_VALUE_INT(*arg);
      gtk_drawing_area_size(GTK_DRAWING_AREA(smoothwave->image),
                            smoothwave->width,smoothwave->height);
      gtk_widget_set_usize(GTK_WIDGET(smoothwave->image),
                           smoothwave->width,smoothwave->height);
      break;
    default:
      break;
  }
}

static void gst_smoothwave_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstSmoothWave *smoothwave;

  /* it's not null if we got it, but it might not be ours */
  smoothwave = GST_SMOOTHWAVE(object);

  switch(id) {
    case ARG_WIDTH: {
      GTK_VALUE_INT(*arg) = smoothwave->width;
      break;
    }
    case ARG_HEIGHT: {
      GTK_VALUE_INT(*arg) = smoothwave->height;
      break;
    }
    case ARG_WIDGET: {
      GTK_VALUE_OBJECT(*arg) = GTK_OBJECT(smoothwave->image);
      break;
    }
    default: {
      arg->type = GTK_TYPE_INVALID;
      break;
    }
  }
}



GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("smoothwave");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the smoothwave element */
  factory = gst_elementfactory_new("smoothwave",GST_TYPE_SMOOTHWAVE,
                                   &gst_smoothwave_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
