/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdlib.h>

#include "vumeter.h"


GstElementDetails gst_vumeter_details = {
  "VU Meter",
  "Sink/Meter",
  "Simple volume indicator",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* VuMeter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_VOLUME,
  ARG_VOLUMEPTR,
  ARG_VOLUME_LEFT,
  ARG_VOLUMEPTR_LEFT,
  ARG_VOLUME_RIGHT,
  ARG_VOLUMEPTR_RIGHT,
};


static void gst_vumeter_class_init(GstVuMeterClass *klass);
static void gst_vumeter_init(GstVuMeter *vumeter);

static void gst_vumeter_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_vumeter_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstElementClass *parent_class = NULL;
//static guint gst_vumeter_signals[LAST_SIGNAL] = { 0 };


GtkType
gst_vumeter_get_type(void) {
  static GtkType vumeter_type = 0;

  if (!vumeter_type) {
    static const GtkTypeInfo vumeter_info = {
      "GstVuMeter",
      sizeof(GstVuMeter),
      sizeof(GstVuMeterClass),
      (GtkClassInitFunc)gst_vumeter_class_init,
      (GtkObjectInitFunc)gst_vumeter_init,
      (GtkArgSetFunc)gst_vumeter_set_arg,
      (GtkArgGetFunc)gst_vumeter_get_arg,
      (GtkClassInitFunc)NULL,
    };
    vumeter_type = gtk_type_unique(GST_TYPE_ELEMENT,&vumeter_info);
  }
  return vumeter_type;
}

static void gst_vumeter_class_init(GstVuMeterClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstVuMeter::volume", GTK_TYPE_INT,
                          GTK_ARG_READABLE, ARG_VOLUME);
  gtk_object_add_arg_type("GstVuMeter::volumeptr", GTK_TYPE_POINTER,
                          GTK_ARG_READABLE, ARG_VOLUMEPTR);
  gtk_object_add_arg_type("GstVuMeter::volume_left", GTK_TYPE_INT,
                          GTK_ARG_READABLE, ARG_VOLUME_LEFT);
  gtk_object_add_arg_type("GstVuMeter::volumeptr_left", GTK_TYPE_POINTER,
                          GTK_ARG_READABLE, ARG_VOLUMEPTR_LEFT);
  gtk_object_add_arg_type("GstVuMeter::volume_right", GTK_TYPE_INT,
                          GTK_ARG_READABLE, ARG_VOLUME_RIGHT);
  gtk_object_add_arg_type("GstVuMeter::volumeptr_right", GTK_TYPE_POINTER,
                          GTK_ARG_READABLE, ARG_VOLUMEPTR_RIGHT);

  gtkobject_class->set_arg = gst_vumeter_set_arg;
  gtkobject_class->get_arg = gst_vumeter_get_arg;
}

static void gst_vumeter_init(GstVuMeter *vumeter) {
  vumeter->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(vumeter),vumeter->sinkpad);
  gst_pad_set_chain_function(vumeter->sinkpad,gst_vumeter_chain);
  vumeter->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(vumeter),vumeter->srcpad);

  vumeter->meta = NULL;
  vumeter->volume = 0.0;
}

GstElement *gst_vumeter_new(gchar *name) {
  GstElement *vumeter = GST_ELEMENT(gtk_type_new(GST_TYPE_VUMETER));
  gst_element_set_name(GST_ELEMENT(vumeter),name);
  return vumeter;
}

void gst_vumeter_chain(GstPad *pad,GstBuffer *buf) {
  GstVuMeter *vumeter;
  gint16 *samples;
  gint samplecount,i;
  gint vl = 0,vr = 0;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  vumeter = GST_VUMETER(pad->parent);

  /* first deal with audio metadata */
//  FIXME
//  if (buf->meta) {
//    if (vumeter->meta != NULL) {
//      /* FIXME: need to unref the old metadata so it goes away */
//    }
//    /* we just make a copy of the pointer */
//    vumeter->meta = (MetaAudioRaw *)(buf->data);
//    /* FIXME: now we have to ref the metadata so it does go away */
//  }

  g_return_if_fail(vumeter->meta != NULL);

/* FIXME: this is very stereo s16_le specific!!! */
  samples = (gint16 *)GST_BUFFER_DATA(buf);
  samplecount = GST_BUFFER_SIZE(buf) / 4;

  for (i=0;i<samplecount;i++) {
    vl = MAX(vl,abs(samples[i]));
    vr = MAX(vr,abs(samples[i+1]));
  }

  vumeter->volume = MAX(vl,vr);
  vumeter->volume_left = vl;
  vumeter->volume_right = vr;

//  gst_trace_add_entry(NULL,0,buf,"vumeter: calculated volume");

//  vumeter->volume = max;
//  g_print("current volume is %d\n",vumeter->volume);

  gst_buffer_unref(buf);
}

static void gst_vumeter_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstVuMeter *vumeter;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VUMETER(object));
  vumeter = GST_VUMETER(object);

  switch(id) {
    default:
      break;
  }
}

static void gst_vumeter_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstVuMeter *vumeter;

  /* it's not null if we got it, but it might not be ours */
  vumeter = GST_VUMETER(object);

  switch(id) {
    case ARG_VOLUME:
      GTK_VALUE_INT(*arg) = vumeter->volume;
      break;
    case ARG_VOLUMEPTR:
      GTK_VALUE_POINTER(*arg) = &vumeter->volume;
      break;
    case ARG_VOLUME_LEFT:
      GTK_VALUE_INT(*arg) = vumeter->volume_left;
      break;
    case ARG_VOLUMEPTR_LEFT:
      GTK_VALUE_POINTER(*arg) = &vumeter->volume_left;
      break;
    case ARG_VOLUME_RIGHT:
      GTK_VALUE_INT(*arg) = vumeter->volume_right;
      break;
    case ARG_VOLUMEPTR_RIGHT:
      GTK_VALUE_POINTER(*arg) = &vumeter->volume_right;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}



GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("vumeter");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the vumeter element */
  factory = gst_elementfactory_new("vumeter",GST_TYPE_VUMETER,
                                   &gst_vumeter_details);
  g_return_val_if_fail(factory != NULL,NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}

