/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdlib.h>
#include <string.h>

#include <vorbis/vorbisenc.h>

#include "vorbisenc.h"


void gst_vorbisenc_chain(GstPad *pad,GstBuffer *buf);

extern GstPadTemplate *enc_src_template, *enc_sink_template;

/* elementfactory information */
GstElementDetails vorbisenc_details = {
  "Ogg Vorbis encoder",
  "Filter/Audio/Encoder",
  "Encodes audio in OGG Vorbis format",
  VERSION,
  "Monty <monty@xiph.org>, "\
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

/* VorbisEnc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
};

static void gst_vorbisenc_class_init(VorbisEncClass *klass);
static void gst_vorbisenc_init(VorbisEnc *vorbisenc);

static GstElementClass *parent_class = NULL;
//static guint gst_vorbisenc_signals[LAST_SIGNAL] = { 0 };

GtkType
vorbisenc_get_type(void) {
  static GtkType vorbisenc_type = 0;

  if (!vorbisenc_type) {
    static const GtkTypeInfo vorbisenc_info = {
      "VorbisEnc",
      sizeof(VorbisEnc),
      sizeof(VorbisEncClass),
      (GtkClassInitFunc)gst_vorbisenc_class_init,
      (GtkObjectInitFunc)gst_vorbisenc_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    vorbisenc_type = gtk_type_unique(GST_TYPE_ELEMENT,&vorbisenc_info);
  }
  return vorbisenc_type;
}

static void
gst_vorbisenc_class_init(VorbisEncClass *klass) {
  GstElementClass *gstelement_class;

  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);
}

static void 
gst_vorbisenc_init (VorbisEnc *vorbisenc) 
{
  vorbisenc->sinkpad = gst_pad_new_from_template (enc_sink_template, "sink");
  gst_element_add_pad(GST_ELEMENT(vorbisenc),vorbisenc->sinkpad);
  gst_pad_set_chain_function(vorbisenc->sinkpad,gst_vorbisenc_chain);

  vorbisenc->srcpad = gst_pad_new_from_template (enc_src_template, "src");
  gst_element_add_pad(GST_ELEMENT(vorbisenc),vorbisenc->srcpad);

  /********** Encode setup ************/

  /* choose an encoding mode */
  /* (mode 0: 44kHz stereo uncoupled, roughly 128kbps VBR) */
  vorbis_info_init(&vorbisenc->vi);
  vorbis_encode_init(&vorbisenc->vi, 2, 44100, -1, 160000, -1);

  /* add a comment */
  vorbis_comment_init(&vorbisenc->vc);
  vorbis_comment_add(&vorbisenc->vc,"Track encoded by GStreamer:vorbisenc");

  /* set up the analysis state and auxiliary encoding storage */
  vorbis_analysis_init(&vorbisenc->vd,&vorbisenc->vi);
  vorbis_block_init(&vorbisenc->vd,&vorbisenc->vb);
  
  /* set up our packet->stream encoder */
  /* pick a random serial number; that way we can more likely build
     chained streams just by concatenation */
  srand(time(NULL));
  ogg_stream_init(&vorbisenc->os,rand());

  /* Vorbis streams begin with three headers; the initial header (with
     most of the codec setup parameters) which is mandated by the Ogg
     bitstream spec.  The second header holds any comment fields.  The
     third header holds the bitstream codebook.  We merely need to
     make the headers, then pass them to libvorbis one at a time;
     libvorbis handles the additional Ogg bitstream constraints */

  {
    ogg_packet header;
    ogg_packet header_comm;
    ogg_packet header_code;

    vorbis_analysis_headerout(&vorbisenc->vd,&vorbisenc->vc,&header,&header_comm,&header_code);
    ogg_stream_packetin(&vorbisenc->os,&header); /* automatically placed in its own
					 page */
    ogg_stream_packetin(&vorbisenc->os,&header_comm);
    ogg_stream_packetin(&vorbisenc->os,&header_code);

    /* no need to write out here.  We'll get to that in the main loop */
  }
}

void gst_vorbisenc_chain(GstPad *pad,GstBuffer *buf) {
  VorbisEnc *vorbisenc;
  guchar *data;
  gulong size;
  gulong i;
  GstBuffer *outbuf;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  vorbisenc = GST_VORBISENC(pad->parent);
  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  if(size==0){
    /* end of file.  this can be done implicitly in the mainline,
       but it's easier to see here in non-clever fashion.
       Tell the library we're at end of stream so that it can handle
       the last frame and mark end of stream in the output properly */
    vorbis_analysis_wrote(&vorbisenc->vd,0);

  }else{
    /* data to encode */

    /* expose the buffer to submit data */
    float **buffer=vorbis_analysis_buffer(&vorbisenc->vd,size/4);
      
    /* uninterleave samples */
    for(i=0;i<size/4;i++){
      buffer[0][i]=((((gchar)data[i*4+1])<<8)|
	      (0x00ff&(int)data[i*4]))/32768.f;
      buffer[1][i]=((((gchar)data[i*4+3])<<8)|
	      (0x00ff&(int)data[i*4+2]))/32768.f;
    }
    
    /* tell the library how much we actually submitted */
    vorbis_analysis_wrote(&vorbisenc->vd,i);
  }

  /* vorbis does some data preanalysis, then divvies up blocks for
     more involved (potentially parallel) processing.  Get a single
     block for encoding now */
  while(vorbis_analysis_blockout(&vorbisenc->vd,&vorbisenc->vb)==1){

    /* analysis */
    vorbis_analysis(&vorbisenc->vb,&vorbisenc->op);
      
    /* weld the packet into the bitstream */
    ogg_stream_packetin(&vorbisenc->os,&vorbisenc->op);

    /* write out pages (if any) */
    while(!vorbisenc->eos){
      int result=ogg_stream_pageout(&vorbisenc->os,&vorbisenc->og);
      if(result==0)break;

      outbuf = gst_buffer_new();
      GST_BUFFER_DATA(outbuf) = g_malloc(vorbisenc->og.header_len+vorbisenc->og.body_len);
      GST_BUFFER_SIZE(outbuf) = vorbisenc->og.header_len+vorbisenc->og.body_len;

      memcpy(GST_BUFFER_DATA(outbuf), vorbisenc->og.header, vorbisenc->og.header_len);
      memcpy(GST_BUFFER_DATA(outbuf)+vorbisenc->og.header_len, vorbisenc->og.body, vorbisenc->og.body_len);

      printf("vorbisenc: encoded buffer of %d bytes\n", GST_BUFFER_SIZE(outbuf));

      gst_pad_push(vorbisenc->srcpad, outbuf);

      /* this could be set above, but for illustrative purposes, I do
         it here (to show that vorbis does know where the stream ends) */
	
      if(ogg_page_eos(&vorbisenc->og)) {
        vorbisenc->eos=1;
      }
    }
  }

  if (vorbisenc->eos) {
    /* clean up and exit.  vorbis_info_clear() must be called last */
  
    ogg_stream_clear(&vorbisenc->os);
    vorbis_block_clear(&vorbisenc->vb);
    vorbis_dsp_clear(&vorbisenc->vd);
    vorbis_info_clear(&vorbisenc->vi);
  }

  gst_buffer_unref(buf);
}
